#include <config.h>
#define _COL3

!--- Compute path-ordered product of link variables along a specified path ----
!------(multi-node version not using link variables in ghost sites)------
!
! content of algorithm  (M.Hayakawa, November, 2010)
!
! (1) divide path into the "portions"
!     Here, a portion denotes the maximal set of consecutive links
!     lying on the path which are contained in a computational node.
!
! (2) multiply link variables of links along each portion
!     and give a "portion of wilson line" 
!
! (3) gather all portions of wilson lines to the master node
!
! (4) multiply all portions of wilson lines only on the master node
!     in a proper order to get the final answer for the wilson line
!
! (5) broadcast the answer to all nodes
!
! Futuer subject : 
!   Is there any algorithm with less number of node communication ?
!
! Ver2.0 deletes the node communitation in constructing portion
! of Wilson lines (All nodes has the same number of Wilson line portion,
! but information on Wilson line is carried by only the corresponding node
!
module path_product_class
  use constants_module
!--- module "param_size" was defined in HMCLIB/chrolog_class.F
!  use param_size
  use lattice_class
  use error_class
#ifndef _singlePU
  use mpi
  use mpi2comlib
#endif
  implicit none
  private

  public :: segment_t, profile_t, path_t, path_portion, path_product_t
  public :: new
  public :: delete
  public :: read
  public :: print
  public :: free
  public :: get_num_seg, get_dir, get_len
  public :: run

  ! byte size of complex number
  integer, parameter :: NPREC = 16
!!$  ! debug purpose
!!$  integer :: run_cnt = 0

  type segment_t
    ! direction ( mu < 0 means minus |mu|-direction ) of segment
    integer :: mu
    ! length of segment
    integer :: len
    ! pointer to the next segement_t
    !   (not used in the applications in this class)
    type(segment_t), pointer :: next => NULL()
  end type segment_t

  type profile_t
    ! number of directed segments constituting the path
    integer :: num_seg
    ! property of directed segment (direction and length)
    type(segment_t), allocatable :: seg(:)
  end type profile_t

  type path_t
    ! global coordinates of starting position
    integer :: x(NDIM)
    ! profile_t object
    type(profile_t) :: prf_obj
  end type path_t

  ! consecutive links contained in one of computation nodes
  type path_portion
    ! Wilson line along this portion (result)
    complex(DP) :: wl(COL,COL)
#ifndef _singlePU
    ! node id including this portion
    integer :: nodeid
#endif
    ! pointer to the next portion data
    type(path_portion), pointer :: next => NULL()
  end type path_portion

  type path_product_t
    ! path of interest (shared by all nodes)
    type(path_t) :: path_obj
    ! list of portions of path, which consist of consecutive links
    ! contained in one of computational nodes 
    type(path_portion), pointer :: portion_obj => NULL()
    ! total number of portions
    integer :: num_portion
  end type path_product_t

!-------------- member functions for segment_t --------------
  interface new
    module procedure new_segment_t
    module procedure copy_segment_t
  end interface

  interface delete
    module procedure delete_segment_t
  end interface

  interface free
    module procedure free_segment_t
  end interface

!-------------- member functions for profile_t -----------------
  interface new
    ! allocate memory to seg(:) of profile_t
    module procedure new_profile_t
    module procedure new_profile_t_from_segments
    module procedure copy_profile_t
  end interface

  interface read
    ! read profile information from file
    module procedure read_profile_t
  end interface

  interface delete
    module procedure delete_profile_t
  end interface

  interface print
    module procedure print_profile_t
  end interface

  interface get_num_seg
    module procedure get_num_seg_profile_t
  end interface

  interface get_dir
    module procedure get_dir_profile_t
  end interface

  interface get_len
    module procedure get_len_profile_t
  end interface

  interface copy
  end interface

!-------------- member functions for path_t -----------------
  interface new
    module procedure new_path_info
  end interface

  interface delete
    module procedure delete_path_info
  end interface

  interface read
    ! read path information from file
    module procedure read_path_info
  end interface

  interface print
    module procedure print_path_info
  end interface

  interface get_num_seg
    module procedure get_num_seg_path_t
  end interface

  interface get_dir
    module procedure get_dir_path_t
  end interface

  interface get_len
    module procedure get_len_path_t
  end interface

!-------------- member functions for path_portion -----------------
  interface new
    module procedure new_path_portion
  end interface

  interface delete
    module procedure delete_path_portion
  end interface

  interface free
    ! free memory of portion chains
    module procedure free_path_portion
  end interface

!---------- member functions for path_product_t ----------------------
  interface new
    module procedure new_path_product_t
  end interface

  interface delete
    module procedure delete_path_product_t
  end interface

  interface free
    module procedure free_path_product_t
  end interface

  interface run
    module procedure run_path_product_t
  end interface

  interface global2local
    module procedure pp_global2local
  end interface

  interface make
    ! make portions of Wilson lines
    module procedure pp_make_portions
  end interface

  interface calc
    ! gather all portions of wilson line and calculate wilson line
    module procedure pp_calc
  end interface

contains
!-------------- IMPLEMENTATION -------------------

subroutine new_segment_t( this )
!
! inout :
!   this : segment_t object
!
  implicit none
  type(segment_t), intent(inout) :: this

  call delete( this )

  return
end subroutine new_segment_t


recursive subroutine copy_segment_t( this, rhs )
!
! inout :
!   this : segment_t object
!
! input :
!   rhs  : original segment_t object
!
  implicit none
  type(segment_t), intent(inout) :: this
  type(segment_t), intent(in) :: rhs

  call delete( this )

  if ( associated( rhs%next ) ) then
    allocate( this%next )
    call copy_segment_t( this%next, rhs%next )
  endif

  this%mu  = rhs%mu
  this%len = rhs%len

  return
end subroutine copy_segment_t


subroutine delete_segment_t( this )
!
! inout :
!   this : segment_t object
!
  implicit none
  type(segment_t), intent(inout) :: this

  call free( this%next )

  return
end subroutine delete_segment_t


recursive subroutine free_segment_t( this )
!
! inout :
!   this : pointer to segment_t object
!
  implicit none
  type(segment_t), intent(inout), pointer :: this

  if ( associated( this ) ) then
    call free_segment_t( this%next )
    deallocate( this )
    NULLIFY( this )
  endif

  return
end subroutine free_segment_t


!-------------------------------------------------------

subroutine new_profile_t( this, num_seg )
!
! inout :
!   this : profile_t object
!
! input :
!   num_seg : number of segments constituting the path of your interest
!
  implicit none
  type(profile_t), intent(inout) :: this
  integer, intent(in) :: num_seg
!--------------------------------------
  integer :: is

  call delete( this )

  this%num_seg = num_seg
  allocate( this%seg(num_seg) )

  return
end subroutine new_profile_t


subroutine new_profile_t_from_segments( this, seg, num_seg )
! convert list of segments to profile_t object
!  whose segment members of profile_t does not have any children 
!
! inout :
!   this : profile_t object
!
! input:
!   seg     : list of segments
!   num_seg : number of segments in the list, optional for check
!
  implicit none
  type(profile_t), intent(inout) :: this
  type(segment_t), intent(in), target :: seg
  integer, intent(in), optional :: num_seg
!--------------------------------------------
  type(segment_t), pointer :: seg_p
  integer :: num, j

  num = 0
  seg_p => seg
  do while ( associated( seg_p ) )
    num = num + 1
    seg_p => seg_p%next
  enddo

  if ( present(num_seg) ) then
    if ( num /= num_seg ) then
      call error_stop("incorrect number of segments!")
    endif
  endif

  call new( this, num )
  seg_p => seg
  do j = 1, num
    this%seg(j)%mu  = seg_p%mu
    this%seg(j)%len = seg_p%len
    seg_p => seg_p%next
  enddo

  return
end subroutine new_profile_t_from_segments


subroutine copy_profile_t( this, rhs )
!
! inout :
!   this : profile_t object
!
! inout:
!   rhs : original profile_t object
!
  implicit none
  type(profile_t), intent(inout) :: this
  type(profile_t), intent(in) :: rhs
!---------------------------------------
  integer :: num_seg, is

  num_seg = rhs%num_seg

  call new( this, num_seg )

  do is = 1, num_seg
    call new( this%seg(is), rhs%seg(is) )
  enddo

  return
end subroutine copy_profile_t


subroutine read_profile_t( this, iunit )
!
! output :
!   this : profile_t object
!
! input :
!   iunit : unit id associated to the file including information of path
!
  implicit none
  type(profile_t), intent(out) :: this
  integer, intent(in) :: iunit

  integer :: j, num_seg

  read(iunit, *)
  ! read total number of segments
  read(iunit, *) num_seg

  call new( this, num_seg )

  read(iunit, *)
  ! read information on directed segments
  do j = 1, this%num_seg
    read(iunit, *) this%seg(j)%mu, this%seg(j)%len
  enddo
 
  return
end subroutine read_profile_t


subroutine delete_profile_t( this )
!
! inout :
!   this : profile_t object
!
  implicit none
  type(profile_t), intent(inout) :: this
!------------------------------------------
  integer :: is

  if ( allocated( this%seg ) ) then
    do is = 1, this%num_seg
      call delete( this%seg(is) )
    enddo
    deallocate( this%seg )
  endif

  return
end subroutine delete_profile_t


subroutine print_profile_t( this, iunit )
!
! input :
!   this : profile_t object
!   iunit : unit id associated to the output file (optional)
!
  implicit none
  type(profile_t), intent(in) :: this
  integer, intent(in), optional :: iunit
!------------------------------------------
  integer :: unit_out = 6
  integer :: j

  if ( present(iunit) ) then
    unit_out = iunit
  endif

  if ( nodeid == 0 ) then
    write( unit_out, '("# dir, length" )' )
    do j = 1, this%num_seg
      write( unit_out, '("#", I3, "  ", I4)' ) this%seg(j)%mu, this%seg(j)%len
    enddo
  endif

  return
end subroutine print_profile_t


function get_num_seg_profile_t( this ) result( num_seg )
!
! output :
!   num_seg : number of segments constituting the profile of your interest
!
! input :
!   this : profile_t object
!
  implicit none
  type(profile_t), intent(in) :: this
  integer :: num_seg

  num_seg = this%num_seg

  return
end function get_num_seg_profile_t


function get_dir_profile_t( this, iseg ) result( idir )
!
! output :
!   idir : direction of the segment of the profile of your interest
!
! input :
!   this : profile_t object
!   iseg : id assinged to the segment of your interest
!
  implicit none
  type(profile_t), intent(in) :: this
  integer, intent(in) :: iseg
  integer :: idir

  idir = this%seg(iseg)%mu

  return
end function get_dir_profile_t


function get_len_profile_t( this, iseg ) result( len )
!
! output :
!   len : length of the segment of the profile of your interest
!
! input :
!   this : profile_t object
!   iseg : id assinged to the segment of your interest
!
  implicit none
  type(profile_t), intent(in) :: this
  integer, intent(in) :: iseg
  integer :: len

  len = this%seg(iseg)%len

  return
end function get_len_profile_t


!----------------------------------------------------------------------

subroutine new_path_info( this, num_seg )
!
! inout :
!   this : path_t object
!
! input :
!   num_seg : number of segments constituting the path of your interest
!
  implicit none
  type(path_t), intent(inout) :: this
  integer, intent(in) :: num_seg

  call new( this%prf_obj, num_seg )

  return
end subroutine new_path_info


subroutine delete_path_info( this )
!
! inout :
!   this : path_t object
!
  implicit none
  type(path_t), intent(inout) :: this

  call delete( this%prf_obj )

  return
end subroutine delete_path_info


subroutine read_path_info( this, iunit )
!
! output :
!   this : path_t object
!
! input :
!   iunit : unit id associated to the file including information of path
!
  implicit none
  type(path_t), intent(out) :: this
  integer, intent(in) :: iunit

  integer :: j, num_seg

  read(iunit, *)
  ! read the global coordinates of starting position 
  read(iunit, *) this%x(1), this%x(2), this%x(3), this%x(4)

  call read( this%prf_obj, iunit )
 
  return
end subroutine read_path_info


subroutine print_path_info( this, iunit )
!
! input :
!   this : path_t object
!   iunit : unit id associated to the output file (optional)
!
  implicit none
  type(path_t), intent(in) :: this
  integer, intent(in), optional :: iunit
!------------------------------------------
  integer :: unit_out = 6
  integer :: j

  if ( present(iunit) ) then
    unit_out = iunit
  endif

  if ( nodeid == 0 ) then
    write( unit_out, '("#Starting point : ", 4I4)' ) this%x(1), this%x(2),  &
 &                                                   this%x(3), this%x(4)
  endif

  call print( this%prf_obj, iunit )

  return
end subroutine print_path_info


function get_num_seg_path_t( this ) result( num_seg )
!
! output :
!   num_seg : number of segments constituting the path of your interest
!
! input :
!   this : path_t object
!
  implicit none
  type(path_t), intent(in) :: this
  integer :: num_seg

  num_seg = get_num_seg( this%prf_obj )

  return
end function get_num_seg_path_t


function get_dir_path_t( this, iseg ) result( idir )
!
! output :
!   idir : direction of the segment of the path of your interest
!
! input :
!   this : path_t object
!   iseg : id assinged to the segment of your interest
!
  implicit none
  type(path_t), intent(in) :: this
  integer, intent(in) :: iseg
  integer :: idir

  idir = get_dir( this%prf_obj, iseg )

  return
end function get_dir_path_t


function get_len_path_t( this, iseg ) result( len )
!
! output :
!   len : length of the segment of the path of your interest
!
! input :
!   this : path_t object
!   iseg : id assinged to the segment of your interest
!
  implicit none
  type(path_t), intent(in) :: this
  integer, intent(in) :: iseg
  integer :: len

  len = get_len( this%prf_obj, iseg )

  return
end function get_len_path_t

!-------------------------------------------------------------------------

subroutine new_path_portion( this )
!
! inout:
!   this : path_portion object
!
  implicit none
  type(path_portion), intent(inout) :: this

  call delete( this )

  return
end subroutine new_path_portion


subroutine delete_path_portion( this )
!
! inout:
!   this : path_portion object
!
  implicit none
  type(path_portion), intent(inout) :: this

  call free( this%next )

  return
end subroutine delete_path_portion


recursive subroutine free_path_portion( this )
!
! inout:
!   this : pointer to path_portion object
!
  implicit none
  type(path_portion), intent(inout), pointer :: this

  if ( associated( this ) ) then
    call free_path_portion( this%next )
    deallocate( this )
  endif

  return
end subroutine free_path_portion

!----------------------------------------------------------------------------

subroutine new_path_product_t( this, iunit )
!
! inout :
!   this : path_product_t object
!
! input :
!   iunit : unit id associated to input file (optional)
!
  implicit none
  type(path_product_t), intent(inout) :: this
  integer, intent(in), optional :: iunit

  call free( this )
  this%num_portion = 0

  if ( present(iunit) ) then
    call read( this%path_obj, iunit )
  endif

  return
end subroutine new_path_product_t


subroutine delete_path_product_t( this )
!
! inout :
!   this : path_product_t object
!
  implicit none
  type(path_product_t), intent(inout) :: this

  call free( this )
  this%num_portion = 0

  return
end subroutine delete_path_product_t


subroutine free_path_product_t( this )
!
! inout :
!   this : path_product_t object
!
  implicit none
  type(path_product_t), intent(inout) :: this

  call delete( this%path_obj )
  if ( associated( this%portion_obj ) ) then
    call delete( this%portion_obj )
    deallocate( this%portion_obj )
  endif

  return
end subroutine free_path_product_t


subroutine run_path_product_t( this, wl, ue, uo )
!
! inout :
!   this : path_product_t object
!
! output:
!   wl : result for wilson line
!
! input :
!   ue : even-site links
!   uo : odd-site links
!
  implicit none
  type(path_product_t), intent(inout) :: this
  complex(DP), intent(out) :: wl(COL,COL)
  complex(DP), intent(in) :: ue(COL,COL,0:NTH,0:NZ1,0:NY1,0:NX1,NDIM)
  complex(DP), intent(in) :: uo(COL,COL,0:NTH,0:NZ1,0:NY1,0:NX1,NDIM)

!!$  run_cnt = run_cnt + 1
!!$  write(*,'("run time:", I3)') run_cnt

  call make( this, ue, uo )
  call calc( this, wl )

  return
end subroutine run_path_product_t


subroutine pp_global2local( xlocal, xnode, xglobal )
!
! global coordinates => (local coordinates, node coordinates)
!
! output :
!   xlocal : 4-dimensional local coordinates
!   xnode  : 4-dimensional node coordinates (0,1,2,...)
!            respecting ipsite(:)
!
! input:
!   xglobal : 4-dimensional global coordinates
!
  implicit none
  integer, intent(out) :: xlocal(NDIM)
  integer, intent(out) :: xnode(NDIM)
  integer, intent(in) :: xglobal(NDIM)
!--------------------------------------------------------
  ! number of total sites 
  integer, parameter :: N(NDIM) = RESHAPE((/NTX,NTY,NTZ,NTT/), (/NDIM/)) 
  ! number of sites per node
  integer, parameter :: nsite(NDIM) = RESHAPE((/NX,NY,NZ,NT/), (/NDIM/))
  integer :: y(NDIM), mu

  ! adjust y(mu) so that 1 <= y(mu) <= N(mu)
  do mu = 1, NDIM
    y(mu) = xglobal(mu)
    do while ( y(mu) < 1 )
      y(mu) = y(mu) + N(mu)
    enddo
    do while ( y(mu) > N(mu) )
      y(mu) = y(mu) - N(mu)
    enddo
  enddo ! mu

  do mu = 1, NDIM
    xlocal(mu) = mod( y(mu) - 1, nsite(mu) ) + 1
    xnode(mu)  = ( y(mu) - 1 ) / nsite(mu)
  enddo

  return
end subroutine pp_global2local


subroutine pp_make_portions( this, ue, uo )
!
! make portions of Wilson lines
!
! inout :
!   this : path_product_t object
!
! input :
!   ue : even-site links
!   uo : odd-site links
!
  implicit none
  type(path_product_t), intent(inout), target :: this
  complex(DP), intent(in) :: ue(COL,COL,0:NTH,0:NZ1,0:NY1,0:NX1,NDIM)
  complex(DP), intent(in) :: uo(COL,COL,0:NTH,0:NZ1,0:NY1,0:NX1,NDIM)
!-----------------------------------------------------------------------
  complex(DP) :: ut(COL,COL)
  type(path_portion), pointer :: portion
  integer :: x(NDIM) ! current global coordinates
  integer :: xl(NDIM) ! local coordinates
  integer :: xn(NDIM) ! current node coordinates
  integer :: xn0(NDIM) ! previous node coordinates
  integer :: num_seg, is, mu, len, ilength !, current_id
  integer :: itb, ieoxyz, ic, jc, nu
  integer :: num_portion ! counting number of portions on this node
  integer :: ierr

  num_seg = get_num_seg( this%path_obj )
  ! current_id = 0
  x(:) = this%path_obj%x(:)

  if ( associated( this%portion_obj ) ) then
    call delete( this%portion_obj )
    deallocate( this%portion_obj )
  endif
  allocate( this%portion_obj )
  portion => this%portion_obj
  num_portion = 1
  portion%wl(1,1) = Z1
  portion%wl(2,1) = Z0
#ifdef _COL3
  portion%wl(3,1) = Z0
#endif
  portion%wl(1,2) = Z0
  portion%wl(2,2) = Z1
#ifdef _COL3
  portion%wl(3,2) = Z0
  portion%wl(1,3) = Z0
  portion%wl(2,3) = Z0
  portion%wl(3,3) = Z1
#endif
  ! set initial xn0
  call global2local( xl, xn0, x )
#ifndef _singlePU
  if (       ( xn0(1) == ipsite(1) ) &
       .and. ( xn0(2) == ipsite(2) ) &
       .and. ( xn0(3) == ipsite(3) ) ) then
    portion%nodeid = nodeid
  else
    portion%nodeid = -1
  endif
#endif

  do is = 1, num_seg

    len = get_len( this%path_obj, is )
    mu = get_dir( this%path_obj, is )

    select case ( mu )

      case(1:) ! mu > 1
        do ilength = 1, len

          call global2local( xl, xn, x )

          itb = xl(4) / 2
          ieoxyz = mod( ipeo + xl(1) + xl(2) + xl(3) + xl(4), 2 )

          if (       ( xn(1) == xn0(1) )  &
 &             .and. ( xn(2) == xn0(2) )  &
 &             .and. ( xn(3) == xn0(3) ) ) then

            if (       ( xn(1) == ipsite(1) ) &
                 .and. ( xn(2) == ipsite(2) ) &
                 .and. ( xn(3) == ipsite(3) ) ) then
              ! multiply u
              !   For instance
              !     ( ... x U(n-mu,mu) ) x U(n,mu)
              !     = (previous wl) x U(n,mu) 
              if ( ieoxyz == 0 ) then
                do jc = 1, COL
                do ic = 1, COL
#ifdef _COL3
                  ut(ic,jc) =   portion%wl(ic,1)  &
 &                              *ue(1,jc,itb,xl(3),xl(2),xl(1),mu)  &
 &                            + portion%wl(ic,2)  &
 &                              *ue(2,jc,itb,xl(3),xl(2),xl(1),mu)  &
 &                            + portion%wl(ic,3)  &
 &                              *ue(3,jc,itb,xl(3),xl(2),xl(1),mu)
#else
                  ut(ic,jc) =   portion%wl(ic,1)  &
 &                              *ue(1,jc,itb,xl(3),xl(2),xl(1),mu)  &
 &                            + portion%wl(ic,2)  &
 &                              *ue(2,jc,itb,xl(3),xl(2),xl(1),mu)
#endif
                enddo
                enddo
              else
                do jc = 1, COL
                do ic = 1, COL
#ifdef _COL3
                  ut(ic,jc) =   portion%wl(ic,1)  &
 &                              *uo(1,jc,itb,xl(3),xl(2),xl(1),mu)  &
 &                            + portion%wl(ic,2)  &
 &                              *uo(2,jc,itb,xl(3),xl(2),xl(1),mu)  &
 &                            + portion%wl(ic,3)  &
 &                              *uo(3,jc,itb,xl(3),xl(2),xl(1),mu)
#else
                  ut(ic,jc) =   portion%wl(ic,1)  &
 &                              *uo(1,jc,itb,xl(3),xl(2),xl(1),mu)  &
 &                            + portion%wl(ic,2)  &
 &                              *uo(2,jc,itb,xl(3),xl(2),xl(1),mu)
#endif
                enddo
                enddo
              endif
              portion%wl(:,:) = ut(:,:)
            endif

          else ! current node \= previous node

            num_portion = num_portion + 1
            allocate( portion%next )
            portion => portion%next
#ifndef _singlePU
            portion%nodeid = xn(1) + NDIMX *( xn(2) + NDIMY *xn(3) )
#endif
            if (       ( xn(1) == ipsite(1) ) &
                 .and. ( xn(2) == ipsite(2) ) &
                 .and. ( xn(3) == ipsite(3) ) ) then
              if ( ieoxyz == 0 ) then
                portion%wl(:,:) = ue(:,:,itb,xl(3),xl(2),xl(1),mu)
              else
                portion%wl(:,:) = uo(:,:,itb,xl(3),xl(2),xl(1),mu)
              endif
            else
              portion%wl(1,1) = Z1
              portion%wl(2,1) = Z0
#ifdef _COL3
              portion%wl(3,1) = Z0
#endif
              portion%wl(1,2) = Z0
              portion%wl(2,2) = Z1
#ifdef _COL3
              portion%wl(3,2) = Z0
              portion%wl(1,3) = Z0
              portion%wl(2,3) = Z0
              portion%wl(3,3) = Z1
#endif
            endif
            xn0(1) = xn(1)
            xn0(2) = xn(2)
            xn0(3) = xn(3)

          endif ! "if" for (current node == previous node)

          x(mu) = x(mu) + 1 ! post-increment

        enddo ! ilength

      case(:-1) ! mu < -1
        mu = - mu
        do ilength = 1, len
          x(mu) = x(mu) - 1 ! pre-decrement

          call global2local( xl, xn, x )

          itb = xl(4) / 2
          ieoxyz = mod( ipeo + xl(1) + xl(2) + xl(3) + xl(4), 2 )

          if (       ( xn(1) == xn0(1) )  &
 &             .and. ( xn(2) == xn0(2) )  &
 &             .and. ( xn(3) == xn0(3) ) ) then


            if (       ( xn(1) == ipsite(1) ) &
                 .and. ( xn(2) == ipsite(2) ) &
                 .and. ( xn(3) == ipsite(3) ) ) then
              ! multiply u
              !   For instance
              !   (... U(n+mu,mu)^dag x U(n,mu)^dag) x U(n-mu,mu)^dag
              !   = (previuos wl) x U(n-mu,mu)^dag
              if ( ieoxyz == 0 ) then
                do jc = 1, COL
                do ic = 1, COL
#ifdef _COL3
                  ut(ic,jc) =   portion%wl(ic,1)  &
 &                              *CONJG( ue(jc,1,itb,xl(3),xl(2),xl(1),mu) )  &
 &                            + portion%wl(ic,2)  &
 &                              *CONJG( ue(jc,2,itb,xl(3),xl(2),xl(1),mu) )  &
 &                            + portion%wl(ic,3)  &
 &                              *CONJG( ue(jc,3,itb,xl(3),xl(2),xl(1),mu) )
#else
                  ut(ic,jc) =   portion%wl(ic,1)  &
 &                              *CONJG( ue(jc,1,itb,xl(3),xl(2),xl(1),mu) )  &
 &                            + portion%wl(ic,2)  &
 &                              *CONJG( ue(jc,2,itb,xl(3),xl(2),xl(1),mu) )
#endif
                enddo
                enddo
              else
                do jc = 1, COL
                do ic = 1, COL
#ifdef _COL3
                  ut(ic,jc) =   portion%wl(ic,1)  &
 &                              *CONJG( uo(jc,1,itb,xl(3),xl(2),xl(1),mu) )  &
 &                            + portion%wl(ic,2)  &
 &                              *CONJG( uo(jc,2,itb,xl(3),xl(2),xl(1),mu) )  &
 &                            + portion%wl(ic,3)  &
 &                              *CONJG( uo(jc,3,itb,xl(3),xl(2),xl(1),mu) )
#else
                  ut(ic,jc) =   portion%wl(ic,1)  &
 &                              *CONJG( uo(jc,1,itb,xl(3),xl(2),xl(1),mu) )  &
 &                            + portion%wl(ic,2)  &
 &                              *CONJG( uo(jc,2,itb,xl(3),xl(2),xl(1),mu) )
#endif
                enddo
                enddo
              endif
              portion%wl(:,:) = ut(:,:)
            endif

          else ! current node /= previous node

            num_portion = num_portion + 1
            allocate( portion%next )
            portion => portion%next
#ifndef _singlePU
            portion%nodeid = xn(1) + NDIMX *( xn(2) + NDIMY *xn(3) )
#endif
            if (       ( xn(1) == ipsite(1) ) &
                 .and. ( xn(2) == ipsite(2) ) &
                 .and. ( xn(3) == ipsite(3) ) ) then
              if ( ieoxyz == 0 ) then
                do jc = 1, COL
                do ic = 1, COL
                  portion%wl(ic,jc) = CONJG( ue(jc,ic,itb,xl(3),xl(2),xl(1),mu) )
                enddo
                enddo
              else
                do jc = 1, COL
                do ic = 1, COL
                  portion%wl(ic,jc) = CONJG( uo(jc,ic,itb,xl(3),xl(2),xl(1),mu) )
                enddo
                enddo
              endif
            else
              portion%wl(1,1) = Z1
              portion%wl(2,1) = Z0
#ifdef _COL3
              portion%wl(3,1) = Z0
#endif
              portion%wl(1,2) = Z0
              portion%wl(2,2) = Z1
#ifdef _COL3
              portion%wl(3,2) = Z0
              portion%wl(1,3) = Z0
              portion%wl(2,3) = Z0
              portion%wl(3,3) = Z1
#endif
            endif
            xn0(1) = xn(1)
            xn0(2) = xn(2)
            xn0(3) = xn(3)

          endif ! "if" for (current node == previous node)

        enddo ! ilength

    end select
  enddo ! is : loop over segments

  this%num_portion = num_portion

  return
end subroutine pp_make_portions


subroutine pp_calc( this, wl )
!
! gather all portions of Wilson lines to master node,
! calculate product of them in a proper order,
! and broadcast result to all nodes
!
! inout :
!   this : path_product_t object
!
! output :
!    wl : the result of wilson line along a specified path
!
  implicit none
  type(path_product_t), intent(inout), target :: this
  complex(DP), intent(out) :: wl(COL,COL)
!-------------------------------------------------------
  complex(DP) :: ut(COL,COL)
  complex(DP), allocatable :: awl(:,:,:)
  type(path_portion), pointer :: portion => NULL()
  integer :: id, ic, jc
#ifndef _singlePU
  integer :: status(MPI_STATUS_SIZE), ierr, itag, isize
  ! integer, allocatable :: req(:)
  integer :: is_here, j, npe
 
  itag = ntag ! keep orignal value of ntag
!  allocate( req(this%num_portion) )
  isize = NPREC*COL*COL ! byte size of color matrices
#endif

  allocate( awl(COL,COL,this%num_portion) )

! gather all portions of Wilson lines to master node
#ifdef _singlePU
  portion => this%portion_obj
  id = 0
  do while ( associated(portion) )
    id = id + 1
    awl(:,:,id) = portion%wl(:,:)
    portion => portion%next
  enddo
#else
  portion => this%portion_obj
  do id = 1, this%num_portion
    itag = itag + 1 ! increment tag on all nodes
    is_here = 0

    ! check if I am the node involving this portion
    if ( portion%nodeid == nodeid ) then
      is_here = 1
    endif

    call comlib_barrier ! ensuring the order of multiplication
    ! in case that portion belongs to master node
    if ( ( is_here == 1 ) .and. ( nodeid == 0 ) ) then
      awl(:,:,id) = portion%wl(:,:)
    else
    ! in case that portion belongs to the node other than master node
    ! send the information to master node
      if ( is_here == 1 ) then
!        call MPI_ISEND( portion%wl(1,1), isize, MPI_CHARACTER,  &
!&                       0, itag, MPI_COMM_WORLD, req(id), ierr )
         call MPI_SEND( portion%wl(1,1), isize, MPI_CHARACTER,  &
&                       0, itag, MPI_COMM_WORLD, ierr )
!        call MPI_WAIT( req(id), status, ierr ) 
      else if ( nodeid == 0 ) then
        call MPI_RECV( awl(1,1,id), isize, MPI_CHARACTER,   &
&                      portion%nodeid, itag, MPI_COMM_WORLD, status, ierr )
!        call MPI_WAIT( req(id), status, ierr ) 
      endif
    endif

    portion => portion%next
  enddo ! id
#endif

  NULLIFY(portion)

! calclate product of portions of Wilson lines
  if ( nodeid == 0 ) then
    wl(1,1) = Z1
    wl(2,1) = Z0
#ifdef _COL3
    wl(3,1) = Z0
#endif
    wl(1,2) = Z0
    wl(2,2) = Z1
#ifdef _COL3
    wl(3,2) = Z0
    wl(1,3) = Z0
    wl(2,3) = Z0
    wl(3,3) = Z1
#endif
    do id = 1, this%num_portion
      do jc = 1, COL
      do ic = 1, COL
#ifdef _COL3
        ut(ic,jc) =   wl(ic,1) * awl(1,jc,id)  &
 &                  + wl(ic,2) * awl(2,jc,id)  &
 &                  + wl(ic,3) * awl(3,jc,id)
#else
        ut(ic,jc) =   wl(ic,1) * awl(1,jc,id)  &
 &                  + wl(ic,2) * awl(2,jc,id)
#endif
      enddo
      enddo
      wl(:,:) = ut(:,:)
    enddo ! id
  endif

#ifndef _singlePU
  ! broadcast result to all nodes from master node
  call MPI_BCAST( wl(1,1), isize, MPI_CHARACTER, 0,  &
 &                MPI_COMM_WORLD, ierr )

!  deallocate( req )
#endif

  deallocate( awl )

  return
end subroutine pp_calc


end module path_product_class
