#ifndef _MAIN_GPU_SOLVER_H
#define _MAIN_GPU_SOLVER_H
////////////////////////////////////////////////////////////////////////////////
// Single precision GPU solver
//
// $Id: main_gpu_solver.h,v 1.4 2011/11/22 08:28:52 ishikawa Exp $
////////////////////////////////////////////////////////////////////////////////

#define BREAKDOWN_THRESHOLD (1.0e-16)

//
// Timer names
//

static cudaTimer_t gsendtime;
static cudaTimer_t ysendtime;
static cudaTimer_t yrecvtime;
static cudaTimer_t solvtime;
static cudaTimer_t multtime;

// flop counter
static long long int flop_count;
static long long int flop_mult_count;
static long long int gsend_count;
static long long int ysend_count;
static long long int yrecv_count;

//
// Texture fetching for the link variables.
//
#ifdef _USE_TEXFETCH_
texture<float4,1,cudaReadModeElementType> tex_ueri;
texture<float4,1,cudaReadModeElementType> tex_uori;
texture<float4,1,cudaReadModeElementType> tex_yori;
#endif

#include "cuMyComplex.h"
#include "YMACRO.h"
#include "cuMatVec.h"
#include "cuWqfAlgebra.h"
#include "cu_mult_forw_t.h"
#include "cu_mult_forw_z.h"
#include "cu_mult_forw_y.h"
#include "cu_mult_forw_x.h"
#include "cu_mult_back_t.h"
#include "cu_mult_back_z.h"
#include "cu_mult_back_y.h"
#include "cu_mult_back_x.h"
#include "cu_mult_clover_term.h"
#include "set_sf_boundary_q.h"
#include "assign_mult_gamma5.h"
#include "assign_mult_clover_term.h"
#include "assign_mult_hop.h"
#include "assign_mult_eoprec_wd.h"
#include "assign_inv_mult_eoprec_wd.h"
#include "assign_mult_Qee.h"

//
// global device memory for LINK/CLOVER fields.
//
cuhgvfield *ued,*uod;
#ifdef _CLOVER
cuhclvfield *fclinved;
cuhclvfield *fclinvod;
#endif

extern "C" void 
#ifdef _CLOVER
__NEW_GPU_SOLVER(cuhgvfield *ue,cuhgvfield *uo,int *ieo,
                cuhclvfield *fclinve,cuhclvfield *fclinvo) {
#else
__NEW_GPU_SOLVER(cuhgvfield *ue,cuhgvfield *uo,int *ieo) {
#endif
//
// send LINK/CLOVER data from HOST to GPU
//

#ifndef _singlePU
  printf(" GPU solver does not support Parallel environment (MPI).\n");
  printf(" Program stops.\n");
  exit(-1);
#endif

#ifdef _DEBUG_  
  fflush(stdout);
  printf("# Start NEW GPU Solver: %s \n",__func__);
  printf("# sizeof(cuhqfield)  =%lld\n",sizeof(cuhqfield));
  printf("# sizeof(cuhgvfield) =%lld\n",sizeof(cuhgvfield));
  printf("# sizeof(cuhclvfield) =%lld\n",sizeof(cuhclvfield));
  fflush(stdout);
#endif

//
// Initialize timers.
//
  createTimer(&gsendtime);
  createTimer(&ysendtime);
  createTimer(&yrecvtime);
  createTimer(&solvtime);
  createTimer(&multtime);

  gsend_count=0;
  ysend_count=0;
  yrecv_count=0;
  flop_count=0;
  flop_mult_count=0;


  startTimer(gsendtime);

  CUDA_SAFE_CALL(cudaMalloc((void**)&ued,sizeof(cuhgvfield)));
  CUT_CHECK_ERROR("cudaMalloc ued faild");

  CUDA_SAFE_CALL(cudaMalloc((void**)&uod,sizeof(cuhgvfield)));
  CUT_CHECK_ERROR("cudaMalloc uod faild");

#ifdef _CLOVER
  CUDA_SAFE_CALL(cudaMalloc((void**)&fclinved,sizeof(cuhclvfield)));
  CUT_CHECK_ERROR("cudaMalloc fclinved faild");

  CUDA_SAFE_CALL(cudaMalloc((void**)&fclinvod,sizeof(cuhclvfield)));
  CUT_CHECK_ERROR("cudaMalloc fclinvod faild");
#endif

  CUDA_SAFE_CALL(cudaMemcpy(ued,ue,sizeof(cuhgvfield),cudaMemcpyHostToDevice));
  CUT_CHECK_ERROR("cudaMemcpy ued faild");

  CUDA_SAFE_CALL(cudaMemcpy(uod,uo,sizeof(cuhgvfield),cudaMemcpyHostToDevice));
  CUT_CHECK_ERROR("cudaMemcpy uod faild");

  gsend_count = sizeof(cuhgvfield)*2;

#ifdef _CLOVER
  CUDA_SAFE_CALL(cudaMemcpy(fclinved,fclinve,sizeof(cuhclvfield),cudaMemcpyHostToDevice));
  CUT_CHECK_ERROR("cudaMemcpy fclinved faild");

  CUDA_SAFE_CALL(cudaMemcpy(fclinvod,fclinvo,sizeof(cuhclvfield),cudaMemcpyHostToDevice));
  CUT_CHECK_ERROR("cudaMemcpy fclinvod faild");

  gsend_count += sizeof(cuhclvfield)*2;
#endif

  stopTimer(gsendtime);

}

extern "C" void 
__DELETE_GPU_SOLVER(const int *print_flag) {

  fflush(stdout);
  if ( 0 == _nodeid_ && 1 == *print_flag ) {
    printf("# Start DELETE GPU Solver: %s\n",__func__);
  }
//
// Finalize GPU solver
//
  CUDA_SAFE_CALL(cudaFree(ued));
  CUT_CHECK_ERROR("cudaFree ued faild");

  CUDA_SAFE_CALL(cudaFree(uod));
  CUT_CHECK_ERROR("cudaFree uod faild");

#ifdef _CLOVER
  CUDA_SAFE_CALL(cudaFree(fclinved));
  CUT_CHECK_ERROR("cudaFree fclinved faild");

  CUDA_SAFE_CALL(cudaFree(fclinvod));
  CUT_CHECK_ERROR("cudaFree fclinvod faild");
#endif

  double etime_solv  = ((double)getTimerValue( solvtime))*1.0e-3;
  double etime_gsend = ((double)getTimerValue(gsendtime))*1.0e-3;
  double etime_ysend = ((double)getTimerValue(ysendtime))*1.0e-3;
  double etime_yrecv = ((double)getTimerValue(yrecvtime))*1.0e-3;
  double etime_mult  = ((double)getTimerValue( multtime))*1.0e-3;
  deleteTimer(solvtime);
  deleteTimer(gsendtime);
  deleteTimer(ysendtime);
  deleteTimer(yrecvtime);
  deleteTimer(multtime);
  double gflops      = ((((double)HNSITE)*((double)flop_count     ))/etime_solv)*1.e-9;
  double gflops_mult = ((((double)HNSITE)*((double)flop_mult_count))/etime_mult)*1.e-9;
  long long int tflop_count      = HNSITE*flop_count;
  long long int tflop_mult_count = HNSITE*flop_mult_count;

  if ( 0 == _nodeid_ && 1 == *print_flag ) {
    printf("%% Mult_GFLOPS_SP      : %15.7F\n",gflops_mult);
    printf("%% Mult_FLOP_SP        : %15.7lld\n",tflop_mult_count);
    printf("%% Mult_SP   ETIME(sec): %15.7F\n",etime_mult);
    printf("%% Solve_GFLOPS_SP     : %15.7F\n",gflops);
    printf("%% Solve_FLOP_SP       : %15.7lld\n",tflop_count);
    printf("%% Solve_SP  ETIME(sec): %15.7F\n",etime_solv);
  }

  float gsendsize = (float)(gsend_count)/1024.0/1024.0;
  float ysendsize = (float)(ysend_count)/1024.0/1024.0;
  float yrecvsize = (float)(yrecv_count)/1024.0/1024.0;

  if ( 0 == _nodeid_ && 1 == *print_flag ) {
    printf("%% GPU_GSend ETIME(sec): %15.7F\n",etime_gsend);
    printf("%% GPU_GSend  SIZE(MiB): %15.7F\n",gsendsize);
    printf("%% GPU_GSend  (MiB/sec): %15.7F\n",gsendsize/etime_gsend);
    printf("%% GPU_YSend ETIME(sec): %15.7F\n",etime_ysend);
    printf("%% GPU_YSend  SIZE(MiB): %15.7F\n",ysendsize);
    printf("%% GPU_YSend  (MiB/sec): %15.7F\n",ysendsize/etime_ysend);
    printf("%% GPU_YRecv ETIME(sec): %15.7F\n",etime_yrecv);
    printf("%% GPU_YRecv  SIZE(MiB): %15.7F\n",yrecvsize);
    printf("%% GPU_YRecv  (MiB/sec): %15.7F\n",yrecvsize/etime_yrecv);
  }
  fflush(stdout);
}

extern "C" void 
__S_BICGSTAB_HMC_GPU(const double     *tol,
                           int       *iter,
                     const double   *kappa,
                     const cuhqfield   *ye,
                           cuhqfield  *yde,
                     const int        *ieo)
//////////////////////////////////////////////////////////////
// BiCGStab GPU Solver to solve 
//  v
//  Dee yde = ye
//
//////////////////////////////////////////////////////////////
{

//===========================================================
// host variables
//===========================================================
  int ioe = (*ieo+1)%2;
  int niter = 5000;
  float kappa2 = (float)((*kappa)*(*kappa));
  float tol2 = (float)((*tol)*(*tol));
  float rtmp0;
  cuComplex alp,bet,omg,rho0,rho1,ctmp0,ctmp1;
#ifdef _DEBUG_
  float err;
#endif
  float tol_bk = BREAKDOWN_THRESHOLD;

//===========================================================
// device variables
//===========================================================
  int *ieod,*ioed;
  float *kappa2d;

  cuComplex *alpd,*betd,*omgd;
  cuhqfield *yed,*yded,*red,*rted,*ped,*qed,*ted,*ved;

#if CUDA_VERSION >= 4000
  CUDA_SAFE_CALL(cudaDeviceSynchronize());
#else
  CUDA_SAFE_CALL(cudaThreadSynchronize());
#endif
#ifdef _DEBUG_
  fflush(stdout);
  printf("# Start GPU Solver: %s\n",__func__);
  printf("# kappa=%24.15e\n",*kappa);
  printf("#   tol=%24.15e\n",*tol);
  printf("#  iter=%10d\n",*iter);
  printf("#   ieo=%2d\n",*ieo);
  fflush(stdout);
#endif
  startTimer(solvtime);

//======================================
// allocate device memory
//======================================
  CUDA_SAFE_CALL(cudaMalloc((void**)&ieod,sizeof(int)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&ioed,sizeof(int)));

  CUDA_SAFE_CALL(cudaMalloc((void**)&alpd   ,sizeof(cuComplex)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&betd   ,sizeof(cuComplex)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&omgd   ,sizeof(cuComplex)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&kappa2d,sizeof(float)));

  CUDA_SAFE_CALL(cudaMalloc((void**)&yed ,sizeof(cuhqfield)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&yded,sizeof(cuhqfield)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&red ,sizeof(cuhqfield)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&rted,sizeof(cuhqfield)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&ped ,sizeof(cuhqfield)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&qed ,sizeof(cuhqfield)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&ted ,sizeof(cuhqfield)));
#if _PREC_ITER_ > 0
  CUDA_SAFE_CALL(cudaMalloc((void**)&ved ,sizeof(cuhqfield)));
#endif

//======================================
// Send data from host to device
//======================================
  startTimer(ysendtime);
  assign(*kappa2d,kappa2);

  CUDA_SAFE_CALL(cudaMemcpy(ieod, ieo,sizeof(int),cudaMemcpyHostToDevice));
  CUDA_SAFE_CALL(cudaMemcpy(ioed,&ioe,sizeof(int),cudaMemcpyHostToDevice));
  CUDA_SAFE_CALL(cudaMemcpy(yed,ye,sizeof(cuhqfield),cudaMemcpyHostToDevice));
  ysend_count += sizeof(int)*2 + sizeof(float) + sizeof(cuhqfield);

  stopTimer(ysendtime);

///////////////////////////////////////////////////////////////////////////////
  // set x_0 = 0 (yde=0)
  clear(yded);

  // r  = b ( = y : rhs of equation )
  // rt = r
  // p  = r
  assign( red,yed);
  assign(rted,red);
  assign( ped,red);

  // rho0 = |re|^2
  rtmp0 = abs2(red,ieod);         // rtmpd = |red|^2
  rho0.x = rtmp0;
  rho0.y = 0.0f;
  flop_count += QABS2;

#ifdef _DEBUG_
  err = sqrt(rtmp0);
#if _PREC_ITER_ > 0
  printf("@%5d ERR:%15.7E PREC_ITER:%4d\n",*iter,err,_PREC_ITER_);
#else
  printf("@%5d ERR:%15.7E\n",*iter,err);
#endif
  fflush(stdout);
#endif

  while (*iter < niter) {
#if _PREC_ITER_ > 0
    // v = M p
    // q = D v
#ifdef _CLOVER
    assign_inv_mult_eoprec_wd(kappa2d,ved,ped,ued,uod,ieod,fclinved,fclinvod);
        assign_mult_eoprec_wd(kappa2d,qed,ved,ued,uod,ieod,fclinved,fclinvod);
#else
    assign_inv_mult_eoprec_wd(kappa2d,ved,ped,ued,uod,ieod);
        assign_mult_eoprec_wd(kappa2d,qed,ved,ued,uod,ieod);
#endif
    flop_count += MULT_MB_PRE*(1+_PREC_ITER_);
#else
    // q = D p
#ifdef _CLOVER
    assign_mult_eoprec_wd(kappa2d,qed,ped,ued,uod,ieod,fclinved,fclinvod);
#else
    assign_mult_eoprec_wd(kappa2d,qed,ped,ued,uod,ieod);
#endif
    flop_count += MULT_MB_PRE;
#endif
    *iter = *iter + 1;

    // alp = rho0/<rt|q>
    ctmp0 = prod(rted,qed,ieod);
    if ( abs(ctmp0) < tol_bk ) {
      printf("# %s :iter= %d: breakdown by <rt|q> = 0\n",__func__,*iter);
      break;
    }
    alp = assign_div(rho0,ctmp0); // alp  = rho0/ctmp0
    assign(*alpd,alp);
    flop_count += QPROD;

#if _PREC_ITER_ > 0
    // x = x + alp v
    // r = r - alp q
    accum_add_mult(yded,alpd,ved);
    accum_sub_mult( red,alpd,qed);
#else
    // x = x + alp p
    // r = r - alp q
    accum_add_mult(yded,alpd,ped);
    accum_sub_mult( red,alpd,qed);
#endif
    flop_count += QMULT1C_ACCUM_ADD*2;

    rtmp0 = abs2(red,ieod);
#ifdef _DEBUG_
    err = sqrt(rtmp0);
    printf("@%5d ERR:%15.7E\n",*iter,err);
    fflush(stdout);
#endif
    flop_count += QABS2;
    if (rtmp0 < tol2) break;

#if _PREC_ITER_ > 0
    // v = M r
    // t = D v
#ifdef _CLOVER
    assign_inv_mult_eoprec_wd(kappa2d,ved,red,ued,uod,ieod,fclinved,fclinvod);
        assign_mult_eoprec_wd(kappa2d,ted,ved,ued,uod,ieod,fclinved,fclinvod);
#else
    assign_inv_mult_eoprec_wd(kappa2d,ved,red,ued,uod,ieod);
        assign_mult_eoprec_wd(kappa2d,ted,ved,ued,uod,ieod);
#endif
    flop_count += MULT_MB_PRE*(1+_PREC_ITER_);
#else
    // t = D r
#ifdef _CLOVER
    assign_mult_eoprec_wd(kappa2d,ted,red,ued,uod,ieod,fclinved,fclinvod);
#else
    assign_mult_eoprec_wd(kappa2d,ted,red,ued,uod,ieod);
#endif
    flop_count += MULT_MB_PRE;
#endif
    *iter = *iter + 1;

    // omg = <t|r>/<t|t>
    rtmp0 = abs2(ted,ieod);
    if ( abs(rtmp0) < tol_bk ) {
      printf("# %s :iter= %d: breakdown by <t|t> = 0\n",__func__,*iter);
      break;
    }
    ctmp0 = prod(ted,red,ieod);
    omg = assign_div(ctmp0,rtmp0);
    if ( abs(omg) < tol_bk ) {
      printf("# %s :iter= %d: breakdown by omg = 0\n",__func__,*iter);
      break;
    }
    assign(*omgd,omg);
    flop_count += QABS2 + QPROD;

#if _PREC_ITER_ > 0
    // x = x + omg v
    // r = r - omg t
    accum_add_mult(yded,omgd,ved);
    accum_sub_mult( red,omgd,ted);
#else
    // x = x + omg r
    // r = r - omg t
    accum_add_mult(yded,omgd,red);
    accum_sub_mult( red,omgd,ted);
#endif
    flop_count += QMULT1C_ACCUM_ADD*2;

    rtmp0 = abs2(red,ieod);
#ifdef _DEBUG_
    err = sqrt(rtmp0);
    printf("@%5d ERR:%15.7E\n",*iter,err);
    fflush(stdout);
#endif
    flop_count += QABS2;
    if (rtmp0 < tol2) break;

    if ( abs(rho0) < tol_bk ) {
      printf("# %s :iter= %d: breakdown by rho0 = 0\n",__func__,*iter);
      break;
    }

    // rho1 = <rt|r>
    rho1 = prod(rted,red,ieod);
    flop_count += QPROD;

    // bet = (alp/omg)*(rho1/rho0)
    ctmp0 = assign_div(alp,omg);
    ctmp1 = assign_div(rho1,rho0);
    bet   = assign_mul(ctmp0,ctmp1);
    assign(*betd,bet);

    // rho0 = rho1;
    rho0 = rho1;

    // p = r + bet ( p - omg q)
    accum_sub_mult(ped,omgd,qed);
    accum_mult_add(ped,betd,red);
    flop_count += QMULT1C_ACCUM_ADD+QMULT2C_ACCUM_ADD;
  } // end while

#ifdef _DEBUG_
///////////////////////////////////
// Check true residual
///////////////////////////////////
    // t = D yd
#ifdef _CLOVER
    assign_mult_eoprec_wd(kappa2d,ted,yded,ued,uod,ieod,fclinved,fclinvod);
#else
    assign_mult_eoprec_wd(kappa2d,ted,yded,ued,uod,ieod);
#endif
    assign(red,yed);
    // r = r - t
    accum_sub(red,ted,ieod);
    rtmp0 = abs2(red,ieod);
    err = sqrt(rtmp0);
    printf("%%%5d ERR:%15.7E\n",*iter,err);
    fflush(stdout);
#endif

//======================================
// receive result from device to host
//======================================
  startTimer(yrecvtime);

  CUDA_SAFE_CALL(cudaMemcpy(yde,yded,sizeof(cuhqfield),cudaMemcpyDeviceToHost));
  yrecv_count += sizeof(cuhqfield);

  stopTimer(yrecvtime);

//======================================
// deallocate device memory
//======================================
  CUDA_SAFE_CALL(cudaFree(ieod));
  CUDA_SAFE_CALL(cudaFree(ioed));

  CUDA_SAFE_CALL(cudaFree(alpd));
  CUDA_SAFE_CALL(cudaFree(betd));
  CUDA_SAFE_CALL(cudaFree(omgd));
  CUDA_SAFE_CALL(cudaFree(kappa2d));

  CUDA_SAFE_CALL(cudaFree(yed));
  CUDA_SAFE_CALL(cudaFree(yded));

  CUDA_SAFE_CALL(cudaFree(red));
  CUDA_SAFE_CALL(cudaFree(rted));
  CUDA_SAFE_CALL(cudaFree(ped));
  CUDA_SAFE_CALL(cudaFree(qed));
  CUDA_SAFE_CALL(cudaFree(ted));
#if _PREC_ITER_ > 0
  CUDA_SAFE_CALL(cudaFree(ved));
#endif

#ifdef _DEBUG_
  fflush(stdout);
#endif
  stopTimer(solvtime);
}

extern "C" void 
__TEST_MULT(double *kappa, cuhqfield *ye, cuhqfield *dye, int *ieo)
{
  int ioe;
  int *ieod, *ioed;
  float *kappa2d, kappa2;
  cuhqfield *yed,*dyed;

  ioe = 1-*ieo;
  kappa2 = float((*kappa)*(*kappa));
  
  CUDA_SAFE_CALL(cudaMalloc((void**)&ieod,sizeof(int)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&ioed,sizeof(int)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&kappa2d,sizeof(float)));
  CUDA_SAFE_CALL(cudaMalloc((void**) &yed,sizeof(cuhqfield)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&dyed,sizeof(cuhqfield)));

  assign(*kappa2d,kappa2);

  CUDA_SAFE_CALL(cudaMemcpy(ieod, ieo,sizeof(int),cudaMemcpyHostToDevice));
  CUDA_SAFE_CALL(cudaMemcpy(ioed,&ioe,sizeof(int),cudaMemcpyHostToDevice));
  CUDA_SAFE_CALL(cudaMemcpy(yed,ye,sizeof(cuhqfield),cudaMemcpyHostToDevice));

  // dye = D ye
#ifdef _CLOVER
  assign_mult_eoprec_wd(kappa2d,dyed,yed,ued,uod,ieod,fclinved,fclinvod);
#else
  assign_mult_eoprec_wd(kappa2d,dyed,yed,ued,uod,ieod);
#endif

  CUDA_SAFE_CALL(cudaMemcpy(dye,dyed,sizeof(cuhqfield),cudaMemcpyDeviceToHost));

  CUDA_SAFE_CALL(cudaFree(dyed));
  CUDA_SAFE_CALL(cudaFree(yed));
  CUDA_SAFE_CALL(cudaFree(kappa2d));
  CUDA_SAFE_CALL(cudaFree(ioed));
  CUDA_SAFE_CALL(cudaFree(ieod));
}


extern "C" void 
__S_CG_HMC_GPU(const double         *tol, 
                     int           *iter,
               const double       *kappa,
               const cuhqfield       *ye,
                     cuhqfield      *yde,
               const int            *ieo,
               const double       *shift,
               const int       *do_shift)
//////////////////////////////////////////////////
// CG GPU Solver to solve
// 
// (Qee - shift) yde = ye
//
//////////////////////////////////////////////////
{

//===========================================================
// host variables
//===========================================================
  int ioe = (*ieo+1)%2;
  int niter = 5000;
  float kappa2 = (float)((*kappa)*(*kappa));
  float   tol2 = (float)((*tol)*(*tol));
  float sshift = (float)(*shift);
  float alp,bet,rho0,rho1,rtmp0;
  cuComplex ctmp0;
#ifdef _DEBUG_
  float err;
#endif
  float tol_bk = BREAKDOWN_THRESHOLD;

//===========================================================
// device variables
//===========================================================
  int   *ieod,*ioed;
  float *kappa2d;
  float *shiftd;
  float *alpd,*betd;

  cuhqfield *yed,*yded,*red,*ped,*qed;

#if CUDA_VERSION >= 4000
  CUDA_SAFE_CALL(cudaDeviceSynchronize());
#else
  CUDA_SAFE_CALL(cudaThreadSynchronize());
#endif
#ifdef _DEBUG_
  fflush(stdout);
  printf("# Start GPU(CG) Solver: %s\n",__func__);
  printf("# kappa=%24.15e\n",*kappa);
  printf("# shift=%24.15e\n",*shift);
  printf("#   tol=%24.15e\n",*tol);
  printf("#  iter=%10d\n",*iter);
  printf("#   ieo=%2d\n",*ieo);
  if ( 1 == *do_shift ) {
    printf("# CGGPU-shift optional, With shift.");
  } else {
    printf("# CGGPU-shift optional, Without shift.");
  }
  fflush(stdout);
#endif
  startTimer(solvtime);


//======================================
// allocate device memory
//======================================
  CUDA_SAFE_CALL(cudaMalloc((void**)&ieod,sizeof(int)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&ioed,sizeof(int)));

  CUDA_SAFE_CALL(cudaMalloc((void**)&alpd   ,sizeof(float)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&betd   ,sizeof(float)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&kappa2d,sizeof(float)));

  if ( 1 == *do_shift ) CUDA_SAFE_CALL(cudaMalloc((void**)&shiftd ,sizeof(float)));

  CUDA_SAFE_CALL(cudaMalloc((void**)&yed ,sizeof(cuhqfield)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&yded,sizeof(cuhqfield)));

  CUDA_SAFE_CALL(cudaMalloc((void**)&red ,sizeof(cuhqfield)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&ped ,sizeof(cuhqfield)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&qed ,sizeof(cuhqfield)));

//======================================
// Send data from host to device
//======================================
  startTimer(ysendtime);
  assign(*kappa2d,kappa2);

  if ( 1 == *do_shift ) assign(*shiftd ,sshift);

  CUDA_SAFE_CALL(cudaMemcpy(ieod, ieo,sizeof(int)      ,cudaMemcpyHostToDevice));
  CUDA_SAFE_CALL(cudaMemcpy(ioed,&ioe,sizeof(int)      ,cudaMemcpyHostToDevice));
  CUDA_SAFE_CALL(cudaMemcpy( yed,  ye,sizeof(cuhqfield),cudaMemcpyHostToDevice));
  ysend_count += sizeof(int)*2 + sizeof(float) + sizeof(cuhqfield);

  stopTimer(ysendtime);

///////////////////////////////////////////////////////////////////////////////
  // set x = 0 (yde=0)
  clear(yded);

  // r  = b ( = y : rhs of equation )
  // p  = r
  assign(red,yed);
  assign(ped,red);

  // rho0 = |r|^2
  rtmp0 = abs2(red,ieod);
  flop_count += QABS2;
  rho0 = rtmp0;

#ifdef _DEBUG_
  err = sqrt(rtmp0);
  printf("@%5d ERR:%15.7E\n",*iter,err);
  fflush(stdout);
#endif

  while (*iter < niter) {

    // q = Qee p
#ifdef _CLOVER
    assign_mult_Qee(kappa2d,qed,ped,ued,uod,ieod,fclinved,fclinvod);
#else
    assign_mult_Qee(kappa2d,qed,ped,ued,uod,ieod);
#endif
    flop_count += MULT_MB_PRE*2;
    if ( 1 == *do_shift ) {
      // q = q - shift * p 
      accum_sub_mult(qed,shiftd,ped);
      flop_count += QACCUM_SUB_MULTR;
    }
    *iter = *iter + 1;

    // alp = rho0/<p|q>
    ctmp0 = prod(ped,qed,ieod);
    rtmp0 = ctmp0.x;
    if ( abs(rtmp0) < tol_bk ) {
      printf("# %s :iter= %d: breakdown by <p|q> = 0\n",__func__,*iter);
      break;
    }
    alp = rho0/rtmp0;
    assign(*alpd,alp);
    flop_count += QABS2;

    // x = x + alp p
    // r = r - alp q
    accum_add_mult(yded,alpd,ped);
    accum_sub_mult( red,alpd,qed);
    flop_count += QACCUM_ADD_MULTR + QACCUM_SUB_MULTR;

    rtmp0 = abs2(red,ieod);
    flop_count += QABS2;
#ifdef _DEBUG_
    err = sqrt(rtmp0);
    printf("@%5d ERR:%15.7E\n",*iter,err);
    fflush(stdout);
#endif
    if (rtmp0 < tol2) break;

    if ( abs(rho0) < tol_bk ) {
      printf("# %s :iter= %d: breakdown by rho0 = 0\n",__func__,*iter);
      break;
    }
    rho1 = rtmp0;
    bet = rho1/rho0;
    rho0 = rho1;
    assign(*betd,bet);

    // p = bet p + r
    accum_mult_add(ped,betd,red);
    flop_count += QACCUM_MULTR_ADD;
  } // end while

#ifdef _DEBUG_
///////////////////////////////////
// Check true residual
///////////////////////////////////
    // q = Qee yd
#ifdef _CLOVER
    assign_mult_Qee(kappa2d,qed,yded,ued,uod,ieod,fclinved,fclinvod);
#else
    assign_mult_Qee(kappa2d,qed,yded,ued,uod,ieod);
#endif
    if ( 1 == *do_shift ) {
      // q = q - shift * yd = (Qee - shift) yd
      accum_sub_mult(qed,shiftd,yded);
      flop_count += QACCUM_SUB_MULTR;
    }
    flop_count += MULT_MB_PRE*2;
    assign(red,yed);
    // r = r - q
    accum_sub(red,qed,ieod);
    flop_count += QACCUM_SUB;
    rtmp0 = abs2(red,ieod);
    err = sqrt(rtmp0);
    printf("%%%5d ERR:%15.7E\n",*iter,err);
    fflush(stdout);
#endif

//======================================
// receive result from device to host
//======================================
  startTimer(yrecvtime);
  CUDA_SAFE_CALL(cudaMemcpy(yde,yded,sizeof(cuhqfield),cudaMemcpyDeviceToHost));
  yrecv_count += sizeof(cuhqfield);
  stopTimer(yrecvtime);

//======================================
// deallocate device memory
//======================================
  CUDA_SAFE_CALL(cudaFree(ieod));
  CUDA_SAFE_CALL(cudaFree(ioed));

  CUDA_SAFE_CALL(cudaFree(alpd));
  CUDA_SAFE_CALL(cudaFree(betd));
  CUDA_SAFE_CALL(cudaFree(kappa2d));

  if ( 1 == *do_shift ) CUDA_SAFE_CALL(cudaFree(shiftd));

  CUDA_SAFE_CALL(cudaFree(yed));
  CUDA_SAFE_CALL(cudaFree(yded));

  CUDA_SAFE_CALL(cudaFree(red));
  CUDA_SAFE_CALL(cudaFree(ped));
  CUDA_SAFE_CALL(cudaFree(qed));

#ifdef _DEBUG_
  fflush(stdout);
#endif
  stopTimer(solvtime);
}

extern "C" void 
__INV_MULT_INV_OEE_CG_GPU(
                      const double         *tol, 
                            int           *iter,
                      const double       *kappa,
                      const cuhqfield       *ye,
                            cuhqfield      *yde,
                      const int            *ieo)
//////////////////////////////////////////////////
// CG GPU Solver to solve
// 
// Oee yde = ye
//
//////////////////////////////////////////////////
{

//===========================================================
// host variables
//===========================================================
  int ioe = (*ieo+1)%2;
  int niter = 5000;
  float kappa2 = (float)((*kappa)*(*kappa));
  float   tol2 = (float)((*tol)*(*tol));
  float alp,bet,rho0,rho1,rtmp0;
  cuComplex ctmp0;
#ifdef _DEBUG_
  float err;
#endif
  float tol_bk = BREAKDOWN_THRESHOLD;

//===========================================================
// device variables
//===========================================================
  int   *ieod,*ioed;
  float *kappa2d;
  float *alpd,*betd;

  cuhqfield *yed,*yded,*red,*ped,*qed;

#if CUDA_VERSION >= 4000
  CUDA_SAFE_CALL(cudaDeviceSynchronize());
#else
  CUDA_SAFE_CALL(cudaThreadSynchronize());
#endif
#ifdef _DEBUG_
  fflush(stdout);
  printf("# Start GPU(CG(1/Oee)) Solver: %s\n",__func__);
  printf("# kappa=%24.15e\n",*kappa);
  printf("#   tol=%24.15e\n",*tol);
  printf("#  iter=%10d\n",*iter);
  printf("#   ieo=%2d\n",*ieo);
  fflush(stdout);
#endif
  startTimer(solvtime);


//======================================
// allocate device memory
//======================================
  CUDA_SAFE_CALL(cudaMalloc((void**)&ieod,sizeof(int)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&ioed,sizeof(int)));

  CUDA_SAFE_CALL(cudaMalloc((void**)&alpd   ,sizeof(float)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&betd   ,sizeof(float)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&kappa2d,sizeof(float)));

  CUDA_SAFE_CALL(cudaMalloc((void**)&yed ,sizeof(cuhqfield)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&yded,sizeof(cuhqfield)));

  CUDA_SAFE_CALL(cudaMalloc((void**)&red ,sizeof(cuhqfield)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&ped ,sizeof(cuhqfield)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&qed ,sizeof(cuhqfield)));

//======================================
// Send data from host to device
//======================================
  startTimer(ysendtime);
  assign(*kappa2d,kappa2);

  CUDA_SAFE_CALL(cudaMemcpy(ieod, ieo,sizeof(int)      ,cudaMemcpyHostToDevice));
  CUDA_SAFE_CALL(cudaMemcpy(ioed,&ioe,sizeof(int)      ,cudaMemcpyHostToDevice));
  CUDA_SAFE_CALL(cudaMemcpy( yed,  ye,sizeof(cuhqfield),cudaMemcpyHostToDevice));
  ysend_count += sizeof(int)*2 + sizeof(float) + sizeof(cuhqfield);

  stopTimer(ysendtime);

///////////////////////////////////////////////////////////////////////////////
  // set x = 0 (yde=0)
  clear(yded);

  // r  = b ( = y : rhs of equation )
  // p  = r
  assign(red,yed);
  assign(ped,red);

  // rho0 = |r|^2
  rtmp0 = abs2(red,ieod);
  flop_count += QABS2;
  rho0 = rtmp0;

#ifdef _DEBUG_
  err = sqrt(rtmp0);
  printf("@%5d ERR:%15.7E\n",*iter,err);
  fflush(stdout);
#endif

  while (*iter < niter) {

    // q = Oee p
#ifdef _CLOVER
    assign_mult_Oee(kappa2d,qed,ped,ued,uod,ieod,fclinved,fclinvod);
#else
    assign_mult_Oee(kappa2d,qed,ped,ued,uod,ieod);
#endif
    flop_count += MULT_MB_PRE*2;
    *iter = *iter + 1;

    // alp = rho0/<p|q>
    ctmp0 = prod(ped,qed,ieod);
    rtmp0 = ctmp0.x;
    if ( abs(rtmp0) < tol_bk ) {
      printf("# %s :iter= %d: breakdown by <p|q> = 0\n",__func__,*iter);
      break;
    }
    alp = rho0/rtmp0;
    assign(*alpd,alp);
    flop_count += QABS2;

    // x = x + alp p
    // r = r - alp q
    accum_add_mult(yded,alpd,ped);
    accum_sub_mult( red,alpd,qed);
    flop_count += QACCUM_ADD_MULTR + QACCUM_SUB_MULTR;

    rtmp0 = abs2(red,ieod);
    flop_count += QABS2;
#ifdef _DEBUG_
    err = sqrt(rtmp0);
    printf("@%5d ERR:%15.7E\n",*iter,err);
    fflush(stdout);
#endif
    if (rtmp0 < tol2) break;

    if ( abs(rho0) < tol_bk ) {
      printf("# %s :iter= %d: breakdown by rho0 = 0\n",__func__,*iter);
      break;
    }
    rho1 = rtmp0;
    bet = rho1/rho0;
    rho0 = rho1;
    assign(*betd,bet);

    // p = bet p + r
    accum_mult_add(ped,betd,red);
    flop_count += QACCUM_MULTR_ADD;
  } // end while

#ifdef _DEBUG_
///////////////////////////////////
// Check true residual
///////////////////////////////////
    // q = Oee yd
#ifdef _CLOVER
    assign_mult_Oee(kappa2d,qed,yded,ued,uod,ieod,fclinved,fclinvod);
#else
    assign_mult_Oee(kappa2d,qed,yded,ued,uod,ieod);
#endif
    flop_count += MULT_MB_PRE*2;
    assign(red,yed);
    // r = r - q
    accum_sub(red,qed,ieod);
    flop_count += QACCUM_SUB;
    rtmp0 = abs2(red,ieod);
    err = sqrt(rtmp0);
    printf("%%%5d ERR:%15.7E\n",*iter,err);
    fflush(stdout);
#endif

//======================================
// receive result from device to host
//======================================
  startTimer(yrecvtime);
  CUDA_SAFE_CALL(cudaMemcpy(yde,yded,sizeof(cuhqfield),cudaMemcpyDeviceToHost));
  yrecv_count += sizeof(cuhqfield);
  stopTimer(yrecvtime);

//======================================
// deallocate device memory
//======================================
  CUDA_SAFE_CALL(cudaFree(ieod));
  CUDA_SAFE_CALL(cudaFree(ioed));

  CUDA_SAFE_CALL(cudaFree(alpd));
  CUDA_SAFE_CALL(cudaFree(betd));
  CUDA_SAFE_CALL(cudaFree(kappa2d));

  CUDA_SAFE_CALL(cudaFree(yed));
  CUDA_SAFE_CALL(cudaFree(yded));

  CUDA_SAFE_CALL(cudaFree(red));
  CUDA_SAFE_CALL(cudaFree(ped));
  CUDA_SAFE_CALL(cudaFree(qed));

#ifdef _DEBUG_
  fflush(stdout);
#endif
  stopTimer(solvtime);
}


#endif

// end of include guard
