#include <config.h>

module hmc_stout_class
!
!= Defines HMC Stout smeared QCD actions
!
!  type(vfield_gluon_wg) :: u
!  type(stout_smearing_parameters) :: stout
!
!  call new(stout,alpha,nstout,type,is_smear_only)
!  - the type (optional) can be         : STOUT_TYPE_ALL (default) or STOUT_TYPE_SPATIAL
!  - the is_smear_only (optional) canbe : .TRUE. or .FALSE. (default)
!
!  call smear(stout,u)
!  - stout%usmr(nstout) is the final smeared link when is_smear_only = .FALSE.
!  - while stout%usmr(1) is the final smeared link when is_smear_only = .TRUE.
!
!  call delete(stout)
!
!
  use lattice_class
  use error_class
  use hmc_status_class
  use field_gauge_class
  use su3exp_class, only : reunit, get_expm_coefs => get_expm_coefs_explicit
!  use su3exp_class, only : get_expm_coefs => get_expm_coefs_taylor_ch
  implicit none
  private
  public :: stout_smearing_parameters
  public :: new
  public :: delete
  public :: read
  public :: print
  public :: print_statistics
  public :: save_config
  public :: force
  public :: smear
  public :: STOUT_TYPE_ALL
  public :: STOUT_TYPE_SPATIAL

  integer, parameter :: STOUT_TYPE_ALL     = 0 ! all link directions are smeared
  integer, parameter :: STOUT_TYPE_SPATIAL = 1 ! saptial links are smeard

  type vfield_gluon_wg_stout_smeared
  !
  ! working array (SU(3) vector field)
  !
    type(vfield_gluon_wg), allocatable :: u
  end type

  type stout_smearing_parameters
  !
  ! stout smearing parameters
  !
    real(DP), private :: alpha  = 0.0_DP
    integer,  private :: nstout = 0
    integer,  private :: type = STOUT_TYPE_ALL
    logical,  private :: is_smear_only = .FALSE.  ! if this is true, all intermediate results are discarded.
    type(vfield_gluon_wg), allocatable :: usmr(:) ! smeard link       (1:nstout)
    type(vfield_gluon_wg), allocatable :: qexp(:) ! smearing exponent (0:nstout-1)
  end type

  interface new
  !
  ! Initialize/create Stout smearing parameters
  !
    module procedure new_stout_params
  end interface

  interface delete
  !
  ! delete Stout smearing parameters
  !
    module procedure delete_stout_params
  end interface

  interface read
  !
  ! read Stout smearing parameters
  !
    module procedure read_stout_params
  end interface

  interface save_config
  !
  ! save Stout smearing parameters
  !
    module procedure save_config_stout_params
  end interface

  interface print
  !
  ! print out Stout smearing parameters
  !
    module procedure print_stout_params
  end interface

  interface print_statistics
  !
  ! print out Stout smearing parameters statistics
  !
    module procedure print_stat_stout_params
  end interface

  interface smear
  !
  ! compute link smearing
  !
    module procedure smear_stout_params
  end interface

  interface force
  !
  ! compute MD force for Stout smeared actions
  !
    module procedure force_stout_params
  end interface

contains

subroutine new_stout_params(this,alpha,nstout,type,is_smear_only)
  implicit none
  type(stout_smearing_parameters), intent(inout) :: this
  real(DP), optional, intent(in) :: alpha
  integer,  optional, intent(in) :: nstout
  integer,  optional, intent(in) :: type
  logical,  optional, intent(in) :: is_smear_only
  character(len=CHARLEN) :: str
  call delete(this)
  if (present(alpha)) then
    this%alpha  = alpha
  else
    this%alpha  = 0.0_DP
  endif
  if (present(nstout)) then
    if ( nstout < 0) then
      write(str,'("new_stout_params: nstout must be >=0 :",A,I4)')__FILE__,__LINE__
      call error_stop(str)
    endif
    this%nstout = nstout
  else
    this%nstout = 0
  endif
  if (present(type)) then
    select case(type)
    case(STOUT_TYPE_ALL)
      this%type = type
    case(STOUT_TYPE_SPATIAL)
      this%type = type
    case default
      write(str,'("new_stout_params: type must be 0 or 1 :",A,I4)')__FILE__,__LINE__
      call error_stop(str)
    end select
  else
    this%type = STOUT_TYPE_ALL
  endif
  if (present(is_smear_only)) then
    this%is_smear_only = is_smear_only
  else
    this%is_smear_only = .FALSE.
  endif
  return
end subroutine

subroutine delete_stout_params(this)
  implicit none
  type(stout_smearing_parameters), intent(inout) :: this
  integer :: i
  if (allocated(this%usmr)) deallocate(this%usmr)
  if (allocated(this%qexp)) deallocate(this%qexp)
  this%alpha  = 0.0_DP
  this%nstout = 0
  return
end subroutine

subroutine read_stout_params(this,iout)
  use comlib
  implicit none
  type(stout_smearing_parameters), intent(inout) :: this
  integer, intent(in) :: iout
  if (0==nodeid) then
    read(iout,*)this%alpha,this%nstout
  endif
#ifndef _singlePU
  call comlib_bcast(this%alpha,0)
  call comlib_bcast(this%nstout,0)
#endif
  return
end subroutine

subroutine print_stout_params(this)
  implicit none
  type(stout_smearing_parameters), intent(in) :: this
  if ( 0 == nodeid ) then
    write(*,'(10X," Stout smearing alpha :",F12.8)') this%alpha
    write(*,'(10X," Stout smearing steps :",I3)')    this%nstout
    select case(this%type)
    case (STOUT_TYPE_ALL)
      write(*,'(10X," Stout smearing type  : ALL LINKS")')
    case (STOUT_TYPE_SPATIAL)
      write(*,'(10X," Stout smearing type  : SPATIAL LINKS ONLY")')
    end select
    if ( this%is_smear_only ) then
      write(*,'(10X," Stout Smearing Only (not for HMC).")')
    else
      write(*,'(10X," Stout Smearing for HMC run.")')
    endif
  endif
  return
end subroutine

subroutine print_stat_stout_params(this)
  implicit none
  type(stout_smearing_parameters), intent(in) :: this
  return
end subroutine

subroutine save_config_stout_params(this,iout)
  implicit none
  type(stout_smearing_parameters), intent(in) :: this
  integer, intent(in) :: iout
  write(iout)this%alpha,this%nstout
  return
end subroutine

subroutine smear_stout_params(this,u)
  implicit none
  type(stout_smearing_parameters), intent(inout) :: this
  type(vfield_gluon_wg),           intent(in)    :: u
  integer :: i,j,k
  character(len=CHARLEN) :: str

  if (0 == this%nstout) return

  if (.not.this%is_smear_only) then

    !-------------------------------------------
    ! For HMC force computation 
    ! All smeared link and exponent
    ! generated in intermediate smearing steps
    ! are hold in the array.
    !-------------------------------------------

    if (.not.allocated(this%usmr)) then
      allocate(this%usmr(1:this%nstout  ))
      do i=1,this%nstout
        call new(this%usmr(i))
      enddo
    endif
    if (.not.allocated(this%qexp)) then
      allocate(this%qexp(0:this%nstout-1))
      do i=0,this%nstout-1
        call new(this%qexp(i))
      enddo
    endif

    if (this%nstout /= SIZE(this%usmr)) then
      write(str,'("Inconsistent Stout working array size u: smear_stout_params: ",A,I3)')__FILE__,__LINE__
      call error_stop(str)
    endif
    if (this%nstout /= SIZE(this%qexp)) then
      write(str,'("Inconsistent Stout working array size q: smear_stout_params: ",A,I3)')__FILE__,__LINE__
      call error_stop(str)
    endif

    !-------------------------------------
    ! start smearing from original link
    !-------------------------------------
    call get_exponent_stout(    this%qexp(0),this%alpha  ,u,this%type)
    call get_smeared_link_stout(this%usmr(1),this%qexp(0),u)

    do i=1,this%nstout-1
      call get_exponent_stout(    this%qexp(i  ),this%alpha,this%usmr(i),this%type)
      call get_smeared_link_stout(this%usmr(i+1),this%qexp(i),this%usmr(i))
    enddo
  

  else

    !-----------------------------------
    ! Smearing only case
    !-----------------------------------
    if (.not.allocated(this%usmr)) then
      allocate(this%usmr(0:1))
      call new(this%usmr(0))
      call new(this%usmr(1))
    endif
    if (.not.allocated(this%qexp)) then
      allocate(this%qexp(0:1))
      call new(this%qexp(0))
      call new(this%qexp(1))
    endif

    !-------------------------------------
    ! start smearing from original link
    !-------------------------------------
    call get_exponent_stout(    this%qexp(0),this%alpha  ,u,this%type)
    call get_smeared_link_stout(this%usmr(1),this%qexp(0),u)

    do i=1,this%nstout-1
      call get_exponent_stout(    this%qexp(0),this%alpha  ,this%usmr(1),this%type)
      call get_smeared_link_stout(this%usmr(0),this%qexp(0),this%usmr(1))
      call assign(this%usmr(1),this%usmr(0))
      call copy_boundary(this%usmr(1))
    enddo

    deallocate(this%qexp)

  endif


  return
end subroutine

subroutine force_stout_params(this,BB,u)
!
! Convert force contribution from those with full Stout smeared link
! to those with un-smeared link using the following iteration steps.
!
! iteration:
!
!  BB_mu(n) = BB_mu(n) exp(iQ^(k)_mu(n)) + Y^(k)_mu(n)
!
! from k=nstout to k=0
!
! this : Stout semaring parameters, smeared links and working vectors.
!   BB : force contribution wtih full Stout meared link (input).
!      : force contribution with un-smeared kink (output)
!    u : un-smeared link
!
  implicit none
  type(stout_smearing_parameters), intent(inout) :: this
  type(vfield_gluon_wg),           intent(inout) :: BB
  type(vfield_gluon_wg),           intent(in)    :: u
  integer :: i
  type(vfield_gluon_wg), allocatable :: Lambda
  character(len=CHARLEN) :: str

  if (0 == this%nstout) return

  if (.not.allocated(this%usmr)) then
    write(str,'("Missing Stout smeared link: force_stout_params: ",A,I3)')__FILE__,__LINE__
    call error_stop(str)
  endif
  if (.not.allocated(this%qexp)) then
    write(str,'("Missing Stout exponent: force_stout_params: ",A,I3)')__FILE__,__LINE__
    call error_stop(str)
  endif
  
  allocate(Lambda)
  call new(Lambda)
  do i=this%nstout-1,1,-1
    call   get_lambda_stout(           Lambda,BB,this%qexp(i),this%usmr(i))
    call update_force_stout(BB,this%alpha,Lambda,this%qexp(i),this%usmr(i))
  enddo
  !------------------------------------
  ! last update uses original link
  !------------------------------------
  call   get_lambda_stout(           Lambda,BB,this%qexp(0),u)
  call update_force_stout(BB,this%alpha,Lambda,this%qexp(0),u)
  deallocate(Lambda)

  return
end subroutine

subroutine get_exponent_stout(q,alpha,u,type)
!
! Q_mu(n) = TracelessHermitianPart[ i * Theta_mu(n)]
!
! Theta_mu(n) = alpha \sum_{nu/=mu} (
!
!    4---<---3     n--->---2     
!    |       |     |       |
!    v       ^  +  ^       v     nu ^
!    |       |     |       |        |
!    n--->---2     6---<---5   )     -->mu
!
!
  implicit none
  type(vfield_gluon_wg), intent(inout) :: q
  real(DP),              intent(in)    :: alpha
  type(vfield_gluon_wg), intent(in)    :: u
  integer,               intent(in)    :: type
  integer :: mu,nu,ieo,ioe
  integer :: ix ,iy ,iz ,ith ,ieoxyz,ic,jc
  integer :: ix2,iy2,iz2,ith2
  integer :: ix4,iy4,iz4,ith4
  integer :: ix5,iy5,iz5,ith5
  integer :: ix6,iy6,iz6,ith6
  complex(DP) :: w234(COL,COL),w1(COL,COL)
  complex(DP) :: w256(COL,COL),th(COL,COL)
  real(DP) :: rcoef,rtmp

  call new(q)

  rcoef = alpha*0.5_DP

  do ieo=0,1
  ioe = 1-ieo
  do mu=1,NDIM

!$OMP PARALLEL DO &
!$OMP& PRIVATE(ix ,iy ,iz ,ith ,ieoxyz,ic,jc,nu,&
!$OMP&         ix2,iy2,iz2,ith2,&
!$OMP&         ix4,iy4,iz4,ith4,&
!$OMP&         ix5,iy5,iz5,ith5,&
!$OMP&         ix6,iy6,iz6,ith6,w234,w256,w1,th,rtmp) _CLPS_
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
      do ith=1-ieoxyz,NTH-ieoxyz

        ix2  = ix  + isx(mu)  ! n + mu
        iy2  = iy  + isy(mu)
        iz2  = iz  + isz(mu)
        ith2 = ith + ieoxyz*ist(mu)

        th(:,:) = Z0

        do nu=1,NDIM
        if (nu == mu) cycle
          ix4  = ix   + isx(nu)  ! n + nu
          iy4  = iy   + isy(nu)
          iz4  = iz   + isz(nu)
          ith4 = ith  + ieoxyz*ist(nu)

          ix5  = ix2  - isx(nu) ! n + mu - nu
          iy5  = iy2  - isy(nu)
          iz5  = iz2  - isz(nu)
          ith5 = ith2 - (1-ieoxyz)*ist(nu)

          ix6  = ix   - isx(nu)  ! n - nu
          iy6  = iy   - isy(nu)
          iz6  = iz   - isz(nu)
          ith6 = ith  - (1-ieoxyz)*ist(nu)

          !----------------------
          !          4---<---3
          !          |       |
          ! th = th+ v       ^
          !          |       |
          !          n       2
          !----------------------
          do jc=1,COL
          do ic=1,COL
            w1(ic,jc) =       u%eo(ioe)%mu(nu)%s(ith2,iz2,iy2,ix2)%u(ic,1)  &
 &                     *conjg(u%eo(ioe)%mu(mu)%s(ith4,iz4,iy4,ix4)%u(jc,1)) &
 &                     +      u%eo(ioe)%mu(nu)%s(ith2,iz2,iy2,ix2)%u(ic,2)  &
 &                     *conjg(u%eo(ioe)%mu(mu)%s(ith4,iz4,iy4,ix4)%u(jc,2)) &
 &                     +      u%eo(ioe)%mu(nu)%s(ith2,iz2,iy2,ix2)%u(ic,3)  &
 &                     *conjg(u%eo(ioe)%mu(mu)%s(ith4,iz4,iy4,ix4)%u(jc,3))
          enddo
          enddo
          do jc=1,COL
          do ic=1,COL
            th(ic,jc) = th(ic,jc)  &
 &                     +w1(ic,1)*conjg(u%eo(ieo)%mu(nu)%s(ith ,iz ,iy ,ix )%u(jc,1))  &
 &                     +w1(ic,2)*conjg(u%eo(ieo)%mu(nu)%s(ith ,iz ,iy ,ix )%u(jc,2))  &
 &                     +w1(ic,3)*conjg(u%eo(ieo)%mu(nu)%s(ith ,iz ,iy ,ix )%u(jc,3))
          enddo
          enddo

          !----------------------
          !          n       2
          !          |       |
          ! th = th+ ^       v
          !          |       |
          !          6---<---5
          !----------------------
          do jc=1,COL
          do ic=1,COL
            w1(ic,jc) = conjg(u%eo(ieo)%mu(nu)%s(ith5,iz5,iy5,ix5)%u(1,ic)) &
 &                     *conjg(u%eo(ioe)%mu(mu)%s(ith6,iz6,iy6,ix6)%u(jc,1)) &
 &                     +conjg(u%eo(ieo)%mu(nu)%s(ith5,iz5,iy5,ix5)%u(2,ic)) &
 &                     *conjg(u%eo(ioe)%mu(mu)%s(ith6,iz6,iy6,ix6)%u(jc,2)) &
 &                     +conjg(u%eo(ieo)%mu(nu)%s(ith5,iz5,iy5,ix5)%u(3,ic)) &
 &                     *conjg(u%eo(ioe)%mu(mu)%s(ith6,iz6,iy6,ix6)%u(jc,3))
          enddo
          enddo
          do jc=1,COL
          do ic=1,COL
            th(ic,jc) = th(ic,jc)  &
 &                     +w1(ic,1)*u%eo(ioe)%mu(nu)%s(ith6,iz6,iy6,ix6)%u(1,jc)  &
 &                     +w1(ic,2)*u%eo(ioe)%mu(nu)%s(ith6,iz6,iy6,ix6)%u(2,jc)  &
 &                     +w1(ic,3)*u%eo(ioe)%mu(nu)%s(ith6,iz6,iy6,ix6)%u(3,jc)
          enddo
          enddo

        enddo ! end of do nu


        !
        !  w1 = u_mu(n) sum_of_staples_mu(n)
        !
        do jc=1,COL
        do ic=1,COL
          w1(ic,jc) = u%eo(ieo)%mu(mu)%s(ith,iz,iy,ix)%u(ic,1)*th(1,jc)  &
 &                   +u%eo(ieo)%mu(mu)%s(ith,iz,iy,ix)%u(ic,2)*th(2,jc)  &
 &                   +u%eo(ieo)%mu(mu)%s(ith,iz,iy,ix)%u(ic,3)*th(3,jc)
        enddo
        enddo

        !
        ! w1 = w1 * i
        !
        do jc=1,COL
        do ic=1,COL
          w1(ic,jc) = CMPLX(-AIMAG(w1(ic,jc)),REAL(w1(ic,jc),kind=DP),kind=DP)
        enddo
        enddo
 
        !
        ! Take Hermitian-traceless part of w1
        !
        ! th = w1 + w1^+ - Tr(w 1 + w1^+)/3
        !
        do jc=1,COL
        do ic=1,COL
          th(ic,jc) = w1(ic,jc) + CONJG(w1(jc,ic))
        enddo
        enddo
        rtmp = ( REAL(th(1,1),kind=DP)   &
 &              +REAL(th(2,2),kind=DP)   &
 &              +REAL(th(3,3),kind=DP) )/COL
        do ic=1,COL
          th(ic,ic) = CMPLX(REAL(th(ic,ic),kind=DP)-rtmp,0.0_DP,kind=DP)
        enddo

        !
        ! Q_mu(n) = alpha*th/2
        !
        do jc=1,COL
        do ic=1,COL
          q%eo(ieo)%mu(mu)%s(ith,iz,iy,ix)%u(ic,jc) = th(ic,jc)*rcoef
        enddo
        enddo

        !
        ! temproal Q_4 is zero for spatial smearing case
        !
        if ( STOUT_TYPE_SPATIAL == type .and. NDIM == mu ) then
          q%eo(ieo)%mu(mu)%s(ith,iz,iy,ix)%u(:,:) = Z0
        endif

      enddo

#ifdef _SF
      !----------------------------------------------------------------
      ! SF boundary links are not smeared.
      !
      ! Q_mu(n) is zero cleared for SF boundary links and ghost sites
      !----------------------------------------------------------------
      if (0 == ieoxyz) then
        q%eo(ieo)%mu(mu)%s(  0,iz,iy,ix)%u(:,:) = Z0   ! itt=0
      endif
      if (1 == ieoxyz .and. NDIM /= mu) then
        q%eo(ieo)%mu(mu)%s(  0,iz,iy,ix)%u(:,:) = Z0   ! itt=1 and mu /= 4
      endif
      if (1 == ieoxyz) then
        q%eo(ieo)%mu(mu)%s(NTH,iz,iy,ix)%u(:,:) = Z0   ! itt=NTT+1
      endif
#endif

    enddo
    enddo
    enddo

  enddo ! end of do mu
  enddo ! end of do ieo

  return
end subroutine

subroutine get_smeared_link_stout(eu,q,u)
!
! eu_mu(n) = exp(i q_mu(n)) * u_mu(n)
!
  implicit none
  type(vfield_gluon_wg), intent(inout) :: eu
  type(vfield_gluon_wg), intent(in)    :: q
  type(vfield_gluon_wg), intent(in)    :: u
  real(DP) :: rnorm
  complex(DP) :: f(COL),d0f(COL),d1f(COL)
  integer :: ieo,ix,iy,iz,ieoxyz,ith,mu,ic,jc
  complex(DP) :: q1(COL,COL),q2(COL,COL),expq(COL,COL)

  call new(eu)

  do ieo=0,1
  do mu=1,NDIM

!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,ith,ic,jc,f,q1,q2,expq) _CLPS_
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
      do ith=1-ieoxyz,NTH-ieoxyz

        q1(:,:) = q%eo(ieo)%mu(mu)%s(ith,iz,iy,ix)%u(:,:)
        q2(:,:) = MATMUL(q1,q1)

        call get_expm_coefs(q1,f)

        do jc=1,COL 
        do ic=1,COL 
          expq(ic,jc) = q2(ic,jc) *f(3) + q1(ic,jc) * f(2)
        enddo
        enddo
        do ic=1,COL 
          expq(ic,ic) = expq(ic,ic) + f(1)
        enddo

        q1(:,:) = MATMUL(expq(:,:),u%eo(ieo)%mu(mu)%s(ith,iz,iy,ix)%u(:,:))
        call reunit(q1)
        eu%eo(ieo)%mu(mu)%s(ith,iz,iy,ix)%u(:,:) = q1(:,:)

      enddo
    enddo
    enddo
    enddo

  enddo ! end of do mu
  enddo ! end of do ieo

  call copy_boundary(eu)

  return
end subroutine

subroutine get_lambda_stout(Lambda,BB,q,u)
!
! Lambda_mu(n) = TrecelessHarmitianPart[X_mu(n)]
!
! X_mu(n) = Q_mu(n)^2 ReTr[R0*U_mu(n)*BB_mu(n)] 
!         + Q_mu(n)   ReTr[R1*U_mu(n)*BB_mu(n)]
!         + ( Q_mu(n)*U_mu(n)*BB_mu(n) + U_mu(n)*BB_mu(n)*Q_mu(n) )*f_2
!         +           U_mu(n)*BB_mu(n) *f_1
!
! R0 = Q_mu(n)^2 (df/dc0)_3 + Q_mu(n) (df/dc0)_2 + I (df/dc0)_1
! R1 = Q_mu(n)^2 (df/dc1)_3 + Q_mu(n) (df/dc1)_2 + I (df/dc1)_1
!
! df/dc0, df/dc1 <= Q_mu(n)
!
  implicit none
  type(vfield_gluon_wg), intent(inout) :: Lambda
  type(vfield_gluon_wg), intent(in)    :: BB
  type(vfield_gluon_wg), intent(in)    :: q
  type(vfield_gluon_wg), intent(in)    :: u
  complex(DP) :: xx(COL,COL)
  complex(DP) :: q1(COL,COL)
  complex(DP) :: q2(COL,COL)
  complex(DP) :: R0(COL,COL)
  complex(DP) :: R1(COL,COL)
  complex(DP) :: ub(COL,COL)
  complex(DP) :: w1(COL,COL)
  complex(DP) :: w2(COL,COL)
  complex(DP) :: f(COL),d0f(COL),d1f(COL)
  real(DP) :: rtmp,tr0,tr1
  integer :: ieo,mu,ix,iy,iz,ieoxyz,ith,ic,jc

  call new(Lambda)

  do ieo=0,1
  do mu=1,NDIM
!$OMP PARALLEL DO &
!$OMP& PRIVATE(ix,iy,iz,ieoxyz,ith,ic,jc,f,d0f,d1f,&
!$OMP&         xx,q1,q2,R0,R1,ub,w1,w2,rtmp,tr0,tr1) _CLPS_
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
      do ith=1-ieoxyz,NTH-ieoxyz

        q1(:,:) = q%eo(ieo)%mu(mu)%s(ith,iz,iy,ix)%u(:,:)
        q2(:,:) = MATMUL(q1,q1)

        !------------------------------------
        ! R0 = Q_mu(n)^2 (df/dc0)_3
        !    + Q_mu(n)   (df/dc0)_2
        !    +      I    (df/dc0)_1
        ! R1 = Q_mu(n)^2 (df/dc1)_3
        !    + Q_mu(n)   (df/dc1)_2
        !    +      I    (df/dc1)_1
        !------------------------------------
        call get_expm_coefs(q1,f,d0f,d1f)
        do jc=1,COL
        do ic=1,COL
          R0(ic,jc) = q2(ic,jc) * d0f(3) + q1(ic,jc) * d0f(2)
          R1(ic,jc) = q2(ic,jc) * d1f(3) + q1(ic,jc) * d1f(2)
        enddo
        enddo
        do ic=1,COL
          R0(ic,ic) = R0(ic,ic) + d0f(1)
          R1(ic,ic) = R1(ic,ic) + d1f(1)
        enddo

        !------------------------------------
        ! ub = U_mu(n) * BB_mu(n)
        !------------------------------------
        do jc=1,COL
        do ic=1,COL
          ub(ic,jc) =  u%eo(ieo)%mu(mu)%s(ith,iz,iy,ix)%u(ic,1) &
 &                   *BB%eo(ieo)%mu(mu)%s(ith,iz,iy,ix)%u(1,jc) &
 &                   + u%eo(ieo)%mu(mu)%s(ith,iz,iy,ix)%u(ic,2) &
 &                   *BB%eo(ieo)%mu(mu)%s(ith,iz,iy,ix)%u(2,jc) &
 &                   + u%eo(ieo)%mu(mu)%s(ith,iz,iy,ix)%u(ic,3) &
 &                   *BB%eo(ieo)%mu(mu)%s(ith,iz,iy,ix)%u(3,jc)
        enddo
        enddo

        !---------------------------------
        ! tr0 = ReTr[R0*U_mu(n)*BB_mu(n)]
        ! tr1 = ReTr[R1*U_mu(n)*BB_mu(n)]
        !---------------------------------
        do ic=1,COL
          w1(ic,ic) = R0(ic,1)*ub(1,ic) &
 &                  + R0(ic,2)*ub(2,ic) &
 &                  + R0(ic,3)*ub(3,ic)
          w2(ic,ic) = R1(ic,1)*ub(1,ic) &
 &                  + R1(ic,2)*ub(2,ic) &
 &                  + R1(ic,3)*ub(3,ic)
        enddo
        tr0 = REAL(w1(1,1),kind=DP)  &
 &           +REAL(w1(2,2),kind=DP)  &
 &           +REAL(w1(3,3),kind=DP)
        tr1 = REAL(w2(1,1),kind=DP)  &
 &           +REAL(w2(2,2),kind=DP)  &
 &           +REAL(w2(3,3),kind=DP)

        !---------------------------------------------
        ! w1 = Q_mu(n)*U_mu(n)*BB_mu(n)
        !    + U_mu(n)*BB_mu(n)*Q_mu(n)
        !---------------------------------------------
        do jc=1,COL
        do ic=1,COL
          w1(ic,jc) = q1(ic,1)*ub(1,jc) &
 &                  + q1(ic,2)*ub(2,jc) &
 &                  + q1(ic,3)*ub(3,jc) &
 &                  + ub(ic,1)*q1(1,jc) &
 &                  + ub(ic,2)*q1(2,jc) &
 &                  + ub(ic,3)*q1(3,jc)
        enddo
        enddo

        !---------------------------------------------
        ! X = q2 tr0 + q1 tr1 + w1*f(3) + ub*f(2)
        !---------------------------------------------
        do jc=1,COL
        do ic=1,COL
          xx(ic,jc) = q2(ic,jc)*tr0   &
 &                  + q1(ic,jc)*tr1   &
 &                  + w1(ic,jc)*f(3)  &
 &                  + ub(ic,jc)*f(2)
        enddo
        enddo

        !---------------------------------------------
        ! Take Hermitian-traceless part of X
        !
        ! Lambda = (X + X^+ - Tr(X + X^+)/3)/2
        !
        !---------------------------------------------
        do jc=1,COL
        do ic=1,COL
          w1(ic,jc) = xx(ic,jc) + CONJG(xx(jc,ic))
        enddo
        enddo
        rtmp = ( REAL(w1(1,1),kind=DP) &
 &              +REAL(w1(2,2),kind=DP) &
 &              +REAL(w1(3,3),kind=DP) )/COL
        do ic=1,COL
          w1(ic,ic) = CMPLX(REAL(w1(ic,ic),kind=DP)-rtmp,0.0_DP,kind=DP)
        enddo

        do jc=1,COL
        do ic=1,COL
          Lambda%eo(ieo)%mu(mu)%s(ith,iz,iy,ix)%u(ic,jc) = w1(ic,jc)*0.5_DP
        enddo
        enddo

      enddo


    enddo
    enddo
    enddo
  enddo ! end of do mu
  enddo ! end of do ieo

  call copy_boundary(Lambda)

#ifdef _SF
  !----------------------------------------------------------------
  ! SF boundary links are not smeared.
  !
  ! set zero at SF boundary in spatial-direction.
  ! set zero on ghost sites
  ! (x,y,z) includes spatial ghost sites.
  !----------------------------------------------------------------
  do ieo=0,1
!$OMP PARALLEL DO &
!$OMP& PRIVATE(ix,iy,iz,ieoxyz,mu) _CLPS3_
    do ix=0,NX1
    do iy=0,NY1
    do iz=0,NZ1
      ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)

      if (0 == ieoxyz) then
        do mu=1,NDIM
          Lambda%eo(ieo)%mu(mu)%s(  0,iz,iy,ix)%u(:,:) = Z0   ! itt=0
        enddo
      endif
      if (1 == ieoxyz) then
        do mu=1,NDIM-1
          Lambda%eo(ieo)%mu(mu)%s(  0,iz,iy,ix)%u(:,:) = Z0   ! itt=1 and spatial direction
        enddo
      endif
      if (1 == ieoxyz) then
        do mu=1,NDIM
          Lambda%eo(ieo)%mu(mu)%s(NTH,iz,iy,ix)%u(:,:) = Z0   ! itt=NTT+1
        enddo
      endif

    enddo
    enddo
    enddo
  enddo ! end of do ieo
#endif

  return
end subroutine

subroutine update_force_stout(BB,alpha,Lambda,q,u)
!
!
! BB_mu(n) = BB_mu(n)*exp(i*Q_mu(n)) + Y_mu(n)
!
! Y_mu(n)  = i*alpha*(
!
!            4---<---3     4---<---3     4---<---3    <4---<---3
!            |       |     |       |     |       |     |       |
!            v       ^  -  v       ^  +  v       ^  -  v       ^
!            |       |     |       |     |       |     |       |
!            n>      2     n       2     n       2     n       2
!                          v                     ^
!
!            n>      2     n       2     n       2     n       2
!            |       |     |       |     |       |     |       |
!         +  ^       v  -  ^       v  +  ^       v  -  ^       v  )
!            |       |     |       |     |       |     |       |
!            6---<---5     6---<---5     6---<---5    <6---<---5
!                                  v     ^                      
!
!  Symbols '>,<','v,^' at site number : Lambda_mu(site_num),Lambda_nu(site_num)
!
!
  implicit none
  type(vfield_gluon_wg), intent(inout) :: BB
  real(DP),              intent(in)    :: alpha
  type(vfield_gluon_wg), intent(in)    :: Lambda
  type(vfield_gluon_wg), intent(in)    :: q
  type(vfield_gluon_wg), intent(in)    :: u
  complex(DP) :: w234(COL,COL)
  complex(DP) :: w561(COL,COL)
  complex(DP) :: w256(COL,COL)
  complex(DP) :: yy(COL,COL)
  complex(DP) :: w1(COL,COL)
  complex(DP) :: w2(COL,COL)
  complex(DP) :: w3(COL,COL)
  complex(DP) :: f(COL)
  integer :: ieo,ioe,mu,nu,ix,iy,iz,ieoxyz,ith,ic,jc
  integer :: ix2,iy2,iz2,ith2
  integer :: ix4,iy4,iz4,ith4
  integer :: ix5,iy5,iz5,ith5
  integer :: ix6,iy6,iz6,ith6

  do ieo=0,1
  ioe=1-ieo
  do mu=1,NDIM
!$OMP PARALLEL DO &
!$OMP& PRIVATE(ix ,iy ,iz ,ith ,ieoxyz,nu,ic,jc,&
!$OMP&         ix2,iy2,iz2,ith2,&
!$OMP&         ix4,iy4,iz4,ith4,&
!$OMP&         ix5,iy5,iz5,ith5,&
!$OMP&         ix6,iy6,iz6,ith6,&
!$OMP&         f,w234,w256,w561,yy,w1,w2,w3) _CLPS_
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
      do ith=1-ieoxyz,NTH-ieoxyz

        ix2  = ix   + isx(mu)  ! n + mu
        iy2  = iy   + isy(mu)
        iz2  = iz   + isz(mu)
        ith2 = ith  + ieoxyz *ist(mu)

        yy(:,:) = Z0

        do nu=1,NDIM
        if (mu == nu) cycle
          ix4  = ix   + isx(nu)   ! n + nu
          iy4  = iy   + isy(nu)
          iz4  = iz   + isz(nu)
          ith4 = ith  + ieoxyz *ist(nu)

          ix5  = ix2  - isx(nu)   ! n + mu - nu
          iy5  = iy2  - isy(nu)
          iz5  = iz2  - isz(nu)
          ith5 = ith2 -(1-ieoxyz)*ist(nu)

          ix6  = ix   - isx(nu)   ! n - nu
          iy6  = iy   - isy(nu)
          iz6  = iz   - isz(nu)
          ith6 = ith  -(1-ieoxyz)*ist(nu)

          do jc=1,COL
          do ic=1,COL
            w234(ic,jc) =       u%eo(ioe)%mu(nu)%s(ith2,iz2,iy2,ix2)%u(ic,1)  &
 &                       *conjg(u%eo(ioe)%mu(mu)%s(ith4,iz4,iy4,ix4)%u(jc,1)) &
 &                      +       u%eo(ioe)%mu(nu)%s(ith2,iz2,iy2,ix2)%u(ic,2)  &
 &                       *conjg(u%eo(ioe)%mu(mu)%s(ith4,iz4,iy4,ix4)%u(jc,2)) &
 &                      +       u%eo(ioe)%mu(nu)%s(ith2,iz2,iy2,ix2)%u(ic,3)  &
 &                       *conjg(u%eo(ioe)%mu(mu)%s(ith4,iz4,iy4,ix4)%u(jc,3))

            w256(ic,jc) = conjg(u%eo(ieo)%mu(nu)%s(ith5,iz5,iy5,ix5)%u(1,ic)) &
 &                       *conjg(u%eo(ioe)%mu(mu)%s(ith6,iz6,iy6,ix6)%u(jc,1)) &
 &                      + conjg(u%eo(ieo)%mu(nu)%s(ith5,iz5,iy5,ix5)%u(2,ic)) &
 &                       *conjg(u%eo(ioe)%mu(mu)%s(ith6,iz6,iy6,ix6)%u(jc,2)) &
 &                      + conjg(u%eo(ieo)%mu(nu)%s(ith5,iz5,iy5,ix5)%u(3,ic)) &
 &                       *conjg(u%eo(ioe)%mu(mu)%s(ith6,iz6,iy6,ix6)%u(jc,3))

            w561(ic,jc) = conjg(u%eo(ioe)%mu(mu)%s(ith6,iz6,iy6,ix6)%u(1,ic)) &
 &                       *      u%eo(ioe)%mu(nu)%s(ith6,iz6,iy6,ix6)%u(1,jc)  &
 &                      + conjg(u%eo(ioe)%mu(mu)%s(ith6,iz6,iy6,ix6)%u(2,ic)) &
 &                       *      u%eo(ioe)%mu(nu)%s(ith6,iz6,iy6,ix6)%u(2,jc)  &
 &                      + conjg(u%eo(ioe)%mu(mu)%s(ith6,iz6,iy6,ix6)%u(3,ic)) &
 &                       *      u%eo(ioe)%mu(nu)%s(ith6,iz6,iy6,ix6)%u(3,jc) 
          enddo
          enddo

          !--------------------------
          !
          !            4---<---3
          !            |       |
          ! yy = yy +  v       ^
          !            |       |
          !            n>      2
          !
          !            4---<---3
          !            |       |
          !         -  v       ^
          !            |       |
          !            n       2
          !            v        
          !--------------------------
          do jc=1,COL
          do ic=1,COL
            w1(ic,jc) = w234(ic,1)*conjg(u%eo(ieo)%mu(nu)%s(ith,iz,iy,ix)%u(jc,1)) &
 &                     +w234(ic,2)*conjg(u%eo(ieo)%mu(nu)%s(ith,iz,iy,ix)%u(jc,2)) &
 &                     +w234(ic,3)*conjg(u%eo(ieo)%mu(nu)%s(ith,iz,iy,ix)%u(jc,3))
          enddo
          enddo
          do jc=1,COL
          do ic=1,COL
            yy(ic,jc) = yy(ic,jc)  &
 &                     +w1(ic,1)*( Lambda%eo(ieo)%mu(mu)%s(ith,iz,iy,ix)%u(1,jc)  &
 &                                -Lambda%eo(ieo)%mu(nu)%s(ith,iz,iy,ix)%u(1,jc)) &
 &                     +w1(ic,2)*( Lambda%eo(ieo)%mu(mu)%s(ith,iz,iy,ix)%u(2,jc)  &
 &                                -Lambda%eo(ieo)%mu(nu)%s(ith,iz,iy,ix)%u(2,jc)) &
 &                     +w1(ic,3)*( Lambda%eo(ieo)%mu(mu)%s(ith,iz,iy,ix)%u(3,jc)  &
 &                                -Lambda%eo(ieo)%mu(nu)%s(ith,iz,iy,ix)%u(3,jc))
          enddo
          enddo

          !--------------------------
          !
          !            4---<---3
          !            |       |
          ! yy = yy +  v       ^
          !            |       |
          !            n       2
          !                    ^
          !
          !           <4---<---3
          !            |       |
          !         -  v       ^
          !            |       |
          !            n       2
          !                     
          !--------------------------
          do jc=1,COL
          do ic=1,COL
            w1(ic,jc) = Lambda%eo(ioe)%mu(nu)%s(ith2,iz2,iy2,ix2)%u(ic,1)*w234(1,jc) &
 &                     +Lambda%eo(ioe)%mu(nu)%s(ith2,iz2,iy2,ix2)%u(ic,2)*w234(2,jc) &
 &                     +Lambda%eo(ioe)%mu(nu)%s(ith2,iz2,iy2,ix2)%u(ic,3)*w234(3,jc) &
 &                     -w234(ic,1)*Lambda%eo(ioe)%mu(mu)%s(ith4,iz4,iy4,ix4)%u(1,jc) &
 &                     -w234(ic,2)*Lambda%eo(ioe)%mu(mu)%s(ith4,iz4,iy4,ix4)%u(2,jc) &
 &                     -w234(ic,3)*Lambda%eo(ioe)%mu(mu)%s(ith4,iz4,iy4,ix4)%u(3,jc)
          enddo
          enddo
          do jc=1,COL
          do ic=1,COL
            yy(ic,jc) = yy(ic,jc)  &
 &                     +w1(ic,1)*conjg(u%eo(ieo)%mu(nu)%s(ith,iz,iy,ix)%u(jc,1)) &
 &                     +w1(ic,2)*conjg(u%eo(ieo)%mu(nu)%s(ith,iz,iy,ix)%u(jc,2)) &
 &                     +w1(ic,3)*conjg(u%eo(ieo)%mu(nu)%s(ith,iz,iy,ix)%u(jc,3))
          enddo
          enddo

          !--------------------------
          !
          !            n>      2
          !            |       |
          ! yy = yy +  ^       v
          !            |       |
          !            6---<---5
          !
          !            n       2
          !            |       |
          !         -  ^       v
          !            |       |
          !            6---<---5
          !                    v
          !--------------------------
          do jc=1,COL
          do ic=1,COL
            w1(ic,jc) = w561(ic,1)*Lambda%eo(ieo)%mu(mu)%s(ith ,iz ,iy ,ix )%u(1,jc) &
 &                     +w561(ic,2)*Lambda%eo(ieo)%mu(mu)%s(ith ,iz ,iy ,ix )%u(2,jc) &
 &                     +w561(ic,3)*Lambda%eo(ieo)%mu(mu)%s(ith ,iz ,iy ,ix )%u(3,jc) &
 &                     -Lambda%eo(ieo)%mu(nu)%s(ith5,iz5,iy5,ix5)%u(ic,1)*w561(1,jc) &
 &                     -Lambda%eo(ieo)%mu(nu)%s(ith5,iz5,iy5,ix5)%u(ic,2)*w561(2,jc) &
 &                     -Lambda%eo(ieo)%mu(nu)%s(ith5,iz5,iy5,ix5)%u(ic,3)*w561(3,jc)
          enddo
          enddo
          do jc=1,COL
          do ic=1,COL
            yy(ic,jc) = yy(ic,jc)  &
 &                    + conjg(u%eo(ieo)%mu(nu)%s(ith5,iz5,iy5,ix5)%u(1,ic))*w1(1,jc) &
 &                    + conjg(u%eo(ieo)%mu(nu)%s(ith5,iz5,iy5,ix5)%u(2,ic))*w1(2,jc) &
 &                    + conjg(u%eo(ieo)%mu(nu)%s(ith5,iz5,iy5,ix5)%u(3,ic))*w1(3,jc)
          enddo
          enddo

          !--------------------------
          !
          !            n       2
          !            |       |
          ! yy = yy +  ^       v
          !            |       |
          !           ^6---<---5
          !
          !            n       2
          !            |       |
          !         -  ^       v
          !            |       |
          !           <6---<---5
          !                     
          !--------------------------
          do jc=1,COL
          do ic=1,COL
            w1(ic,jc) = w256(ic,1)*( Lambda%eo(ioe)%mu(nu)%s(ith6,iz6,iy6,ix6)%u(1,jc)  &
 &                                  -Lambda%eo(ioe)%mu(mu)%s(ith6,iz6,iy6,ix6)%u(1,jc)) &
 &                     +w256(ic,2)*( Lambda%eo(ioe)%mu(nu)%s(ith6,iz6,iy6,ix6)%u(2,jc)  &
 &                                  -Lambda%eo(ioe)%mu(mu)%s(ith6,iz6,iy6,ix6)%u(2,jc)) &
 &                     +w256(ic,3)*( Lambda%eo(ioe)%mu(nu)%s(ith6,iz6,iy6,ix6)%u(3,jc)  &
 &                                  -Lambda%eo(ioe)%mu(mu)%s(ith6,iz6,iy6,ix6)%u(3,jc))
          enddo
          enddo
          do jc=1,COL
          do ic=1,COL
            yy(ic,jc) = yy(ic,jc)  &
 &                    + w1(ic,1)*u%eo(ioe)%mu(nu)%s(ith6,iz6,iy6,ix6)%u(1,jc) &
 &                    + w1(ic,2)*u%eo(ioe)%mu(nu)%s(ith6,iz6,iy6,ix6)%u(2,jc) &
 &                    + w1(ic,3)*u%eo(ioe)%mu(nu)%s(ith6,iz6,iy6,ix6)%u(3,jc)
          enddo
          enddo

        enddo ! end of do nu

        yy(:,:) = ZI*alpha*yy(:,:)

        !-----------------------------
        ! w3 = exp(i Q_mu(n))
        !-----------------------------
        w1(:,:) = q%eo(ieo)%mu(mu)%s(ith,iz,iy,ix)%u(:,:)
        w2(:,:) = MATMUL(w1,w1)
        call get_expm_coefs(w1,f)
        do jc=1,COL
        do ic=1,COL
          w3(ic,jc) = w2(ic,jc) * f(3) + w1(ic,jc) * f(2)
        enddo
        enddo
        do ic=1,COL
          w3(ic,ic) = w3(ic,ic) + f(1)
        enddo

        !--------------------------------------------
        ! w2 = BB_mu(n)*exp(i Q_mu(n)) + Y_mu(n)
        !--------------------------------------------
        w1(:,:) = BB%eo(ieo)%mu(mu)%s(ith,iz,iy,ix)%u(:,:)
        w2(:,:) = MATMUL(w1,w3) + yy(:,:)
        BB%eo(ieo)%mu(mu)%s(ith,iz,iy,ix)%u(:,:) = w2(:,:)

      enddo

#ifdef _SF
      !----------------------------------------------------------------
      ! SF boundary links are not smeared.
      ! force contribution for SF boundary links are zero
      !----------------------------------------------------------------
      if (0 == ieoxyz) then
        BB%eo(ieo)%mu(mu)%s(  0,iz,iy,ix)%u(:,:) = Z0   ! itt=0
      endif
      if (1 == ieoxyz .and. NDIM /= mu) then
        BB%eo(ieo)%mu(mu)%s(  0,iz,iy,ix)%u(:,:) = Z0   ! itt=1 and mu /= 4
      endif
      if (1 == ieoxyz) then
        BB%eo(ieo)%mu(mu)%s(NTH,iz,iy,ix)%u(:,:) = Z0   ! itt=NTT+1
      endif
#endif

    enddo
    enddo
    enddo
  enddo ! end of do mu
  enddo ! end of do ieo

  return
end subroutine

end module
