#include <config.h>

module reweighting_truncd_overlap_class
!
!= Reweighting Truncated-DW-Overlap quarks.
!
  use lattice_class
  use error_class
  use logfile_class
  use hmc_logfile_class
  use solver_parameter_class
  use field_gauge_class
  use quark_dw_ovlp_class
  implicit none
  private

  public :: reweighting_truncd_dw_overlap
  public :: new
  public :: delete
  public :: read
  public :: save_config
  public :: print
  public :: print_statistics
  public :: run
  public :: set_id

  integer, parameter :: MAX_ITER = 100000    ! solver maximum iteration count

  type reweighting_truncd_dw_overlap

    type(quark_truncd_dw_overlap), public, pointer :: target_phys_param => NULL() ! reweighting target parameters
    type(quark_truncd_dw_overlap), public, pointer :: hmc_phys_param    => NULL() ! truncated dw overlap HMC action parameters

    type(solver_parameter), private :: reweighting_inv          ! solver tolerance

    type(hmc_logfile),      private :: hmc_reweighting_solver_log   ! solver log
    type(hmc_logfile),      private :: hmc_reweighting_log          ! Reweighting log
    type(hmc_logfile),      private :: hmc_lowmode_log              ! Kernel lowmode log

    type(logfile),          private :: reweighting_solver_log   ! solver log
    type(logfile),          private :: reweighting_log          ! Reweighting log
    type(logfile),          private :: lowmode_log              ! Kernel lowmode log

    character(CHARLEN),     private :: reweighting_solver_log_fname = "tdwov_reweighting_solver_log."
    character(CHARLEN),     private :: reweighting_log_fname        = "tdwov_reweighting_log."
    character(CHARLEN),     private :: lowmode_log_fname            = "tdwov_reweighting_lowmode_log."

    real(DP), private :: run_time = 0.0_DP
    real(DP), private :: shift_kernel_eval_min = 0.0_DP
    real(DP), private :: shift_kernel_eval_max = 0.0_DP

    integer,                private :: NEV=0                      ! number of low-modes to be improved
    integer,                private :: num_reweighting_steps = 0  ! reweighting step number
    integer,                private :: num_noise_rew         = 0  ! reweighting noize number
    integer,                private :: myid=0                     ! quark number

    logical, private :: is_hmc = .TRUE.

  end type

  interface new
    module procedure new_tdwov_reweighting
  end interface

  interface delete
    module procedure delete_tdwov_reweighting
  end interface

  interface read
    module procedure read_tdwov_reweighting
  end interface

  interface save_config
    module procedure save_config_tdwov_reweighting
  end interface

  interface print_statistics
    module procedure print_stat_tdwov_reweighting
  end interface

  interface print
    module procedure print_tdwov_reweighting
  end interface

  interface run
    module procedure run_tdwov_reweighting
  end interface

  interface set_id
    module procedure set_id_tdwov_reweighting
  end interface

contains 

subroutine set_id_tdwov_reweighting(this,id)
!
! set quark # id
!
  implicit none
  type(reweighting_truncd_dw_overlap), intent(inout) :: this
  integer, intent(in) :: id
  this%myid = id
  call set_id(this%target_phys_param,id)
  call set_id(this%hmc_phys_param,id)
  return
end subroutine

subroutine new_tdwov_reweighting(this,id,is_hmc)
!
! initialize reweighting
!
  implicit none
  type(reweighting_truncd_dw_overlap), intent(inout) :: this
  integer, intent(in) :: id
  character(len=CHARLEN) :: str
  logical, optional, intent(in) :: is_hmc

  this%run_time = 0.0_DP

  this%myid = id
  allocate(this%target_phys_param)
  allocate(this%hmc_phys_param)
  call set_id(this%target_phys_param,id)
  call set_id(this%hmc_phys_param,id)

  this%shift_kernel_eval_min = 0.0_DP
  this%shift_kernel_eval_max = 0.0_DP

  this%NEV = 0
  this%num_reweighting_steps = 0
  this%num_noise_rew = 0

  if (present(is_hmc)) then
    this%is_hmc = is_hmc
  else
    this%is_hmc = .TRUE.
  endif


  if (this%is_hmc) then
    write(str,'(I2.2)')id
  else
    write(str,'(I6.6)')id
  endif

  this%reweighting_solver_log_fname = TRIM(this%reweighting_solver_log_fname)//TRIM(str)
         this%reweighting_log_fname =        TRIM(this%reweighting_log_fname)//TRIM(str)
             this%lowmode_log_fname =            TRIM(this%lowmode_log_fname)//TRIM(str)

  if (this%is_hmc) then
    call new(this%hmc_reweighting_solver_log,TRIM(ADJUSTL(this%reweighting_solver_log_fname)))
    call new(       this%hmc_reweighting_log,       TRIM(ADJUSTL(this%reweighting_log_fname)))
    call new(           this%hmc_lowmode_log,           TRIM(ADJUSTL(this%lowmode_log_fname)))
  else
    write(*,*)" Non HMC logfile Opens"
    call new(this%reweighting_solver_log,TRIM(ADJUSTL(this%reweighting_solver_log_fname)))
    call new(       this%reweighting_log,       TRIM(ADJUSTL(this%reweighting_log_fname)))
    call new(           this%lowmode_log,           TRIM(ADJUSTL(this%lowmode_log_fname)))
  endif

  return
end subroutine

subroutine delete_tdwov_reweighting(this)
!
! \delete reweighting
!
  implicit none
  type(reweighting_truncd_dw_overlap), intent(inout) :: this

  call delete(this%target_phys_param)
  call delete(this%hmc_phys_param)
  deallocate(this%target_phys_param)
  deallocate(this%hmc_phys_param)

  this%target_phys_param => NULL()
  this%hmc_phys_param    => NULL()

  if (this%is_hmc) then
    call delete(this%hmc_reweighting_solver_log)
    call delete(this%hmc_reweighting_log)
    call delete(this%hmc_lowmode_log)
  else
    call delete(this%reweighting_solver_log)
    call delete(this%reweighting_log)
    call delete(this%lowmode_log)
  endif

  return
end subroutine

subroutine read_tdwov_reweighting(this,iout)
!
! \read reweighting job parmeters
!
  use comlib
  implicit none
  type(reweighting_truncd_dw_overlap), intent(inout) :: this
  integer, intent(in) :: iout

  if (nodeid==0) then
    read(iout,*) ! skip comment line
  endif

  call read(this%target_phys_param,iout)

  if (nodeid==0) then
    read(iout,*) ! skip comment line
  endif

  call read(this%hmc_phys_param,iout)

  if (nodeid==0) then
    read(iout,*) ! skip comment line
  endif

  if (nodeid==0) then
    read(iout,*)this%reweighting_inv%tol       ! solver tolerance
    read(iout,*)this%num_noise_rew             ! number of gaussian noise for determinant
    read(iout,*)this%num_reweighting_steps     ! reweighing steps NS
    read(iout,*)this%shift_kernel_eval_min,this%shift_kernel_eval_max  ! min and max estimate for |Hw| (sign kernel)
    read(iout,*)this%NEV                       ! number of low mode eigen vectors to be projected out
  endif
#ifndef _singlePU
  call comlib_bcast(this%reweighting_inv%tol,0)
  call comlib_bcast(this%num_noise_rew,0)
  call comlib_bcast(this%num_reweighting_steps,0)
  call comlib_bcast(this%shift_kernel_eval_min,0)
  call comlib_bcast(this%shift_kernel_eval_max,0)
  call comlib_bcast(this%NEV,0)
#endif

  return
end subroutine

subroutine save_config_tdwov_reweighting(this,iout)
!
! \save reweighting config
!
  implicit none
  type(reweighting_truncd_dw_overlap), intent(in) :: this
  integer,                             intent(in) :: iout   ! configuration file unit id
  return
end subroutine

subroutine print_tdwov_reweighting(this)
!
! \print out reweighting parameters
!
  implicit none
  type(reweighting_truncd_dw_overlap), intent(inout) :: this

  if (nodeid==0) then

    write(*,'("*** Reweighting Target Action paramter")')
    call print(this%target_phys_param)
    write(*,'("*** Reweighting HMC Action paramter")')
    call print(this%hmc_phys_param)
    write(*,'("** Reweighting paramter")')
    write(*,'("        Solver Stopping Condition :",E18.8)')this%reweighting_inv%tol
    write(*,'("                  Number of Noise :",I4)')this%num_noise_rew
    write(*,'("     Number of Rewieighting Steps :",I4)')this%num_reweighting_steps
    write(*,'("              Number of Low-modes :",I4)')this%NEV
    write(*,'(" Sign Kernel eigen value range estimate ")')
    write(*,'("                        min(|Hw|) :",E18.8)')this%shift_kernel_eval_min
    write(*,'("                        max(|Hw|) :",E18.8)')this%shift_kernel_eval_max

  endif

  return
end subroutine

subroutine print_stat_tdwov_reweighting(this)
!
! \print out reweighting statistics
!
  implicit none
  type(reweighting_truncd_dw_overlap), intent(inout) :: this

  if (nodeid==0) then
  write(*,'("**** Quark # ",I3,"  Reweighing Truncated-DW-Overlap")')this%myid
  call print(this)
  write(*,'("             Reweighting Run Time :",F12.4)')this%run_time
  endif

  return
end subroutine

subroutine run_tdwov_reweighting(this,itraj,u)
!
! Compute reweighting factor for quark determinant
!
! Extra dimension size NS is enlarged by reweighting
!
!\[
! W = \exp( - |D_{OV}(target) \ D_{OV}(HMC) \eta|^2 + |\eta|^2)
!\]
! with gaussian noise  $ \eta $ .
!
!
  use timer_class
  use hmc_status_class
  implicit none
  type(reweighting_truncd_dw_overlap),intent(inout) :: this  ! quark action/algorithm parameters
  integer,                intent(in)    :: itraj
  type(vfield_gluon_wg),  intent(in)    :: u
  type(field_quark_wg),  allocatable :: eta,v,w
  type(dwf_low_modes),   allocatable :: LV

  type(quark_truncd_dw_overlap) :: phys_param_ns0
  type(quark_truncd_dw_overlap) :: phys_param_ns1

  integer :: iout,iiter,inoise
  integer :: NS_org,NS_target,dNS,isteps_ns,NS0,NS1
  real(DP), parameter :: EIGEN_TOL = 1.0e-13_DP
  real(DP) :: tol,rtmp0,rtmp1,rtmp2,mq, eval_min
  real(DP) :: ave_dS, ave_exp
  complex(DP) :: ctmp
  character(CHARLEN) :: str
  logical :: is_lowprojd
  type(timer) :: etime
  integer :: iout_reweighting_log
  integer :: iout_solver_log
  integer :: iout_lowmode_log

  call new(etime)
  call tic(etime)

  if (this%NEV > 0) then
    is_lowprojd = .true.
  else
    is_lowprojd = .false.
  endif

  call set_zolotarev(this%target_phys_param)

  NS_org    = get_NS(this%hmc_phys_param)
  NS_target = get_NS(this%target_phys_param)
  dNS = (NS_target - NS_org)/this%num_reweighting_steps

#ifdef _DEBUG
  if (0 == nodeid) then
    write(*,'("NS_org,NS_target,dNS=",3I3)')NS_org,NS_target,dNS
  endif
#endif

  if (this%is_hmc) then
    write(str,'("#HMC mode")')
    call print(this%hmc_lowmode_log,TRIM(str))
    iout_lowmode_log = get_file_unit(this%hmc_lowmode_log)
  else
    write(str,'("#NonHMC mode")')
    call print(this%lowmode_log,TRIM(str))
    iout_lowmode_log = get_file_unit(this%lowmode_log)
  endif

  if (this%is_hmc) then
    write(str,'("#HMC mode")')
    call print(this%hmc_reweighting_solver_log,TRIM(str))
    iout_solver_log = get_file_unit(this%hmc_reweighting_solver_log)
  else
    write(str,'("#NonHMC mode")')
    call print(this%reweighting_solver_log,TRIM(str))
    iout_solver_log = get_file_unit(this%reweighting_solver_log)
  endif

  if (this%is_hmc) then
    write(str,'("#HMC mode")')
    call print(this%hmc_reweighting_log,TRIM(str))
    iout_reweighting_log = get_file_unit(this%hmc_reweighting_log)
  else
    write(str,'("#NonHMC mode")')
    call print(this%reweighting_log,TRIM(str))
    iout_reweighting_log = get_file_unit(this%reweighting_log)
  endif

  if (is_lowprojd) then

    !---------------------------------------------------
    ! overlap kernel lowmode computation
    ! kernel = Hw for Domain-Wall fermion
    !---------------------------------------------------
    allocate(LV)

    call new(LV,this%NEV)
    LV%shift_kernel_eval_min = this%shift_kernel_eval_min
    LV%shift_kernel_eval_max = this%shift_kernel_eval_max

    write(iout_lowmode_log,'("# REWEIGHTING LOWMODE SOLVER, NEV=",I3)') this%NEV

    tol    = EIGEN_TOL
    iiter  = MAX_ITER
    call get_low_modes(LV,this%target_phys_param,iout_lowmode_log,tol,iiter,u)

  else

    !---------------------------------------------------
    ! overlap kernel lowmode computation
    ! kernel = Hw for Domain-Wall fermion
    !---------------------------------------------------
    allocate(LV)

    call new(LV,NEV=6)
    LV%shift_kernel_eval_min = this%shift_kernel_eval_min
    LV%shift_kernel_eval_max = this%shift_kernel_eval_max

    write(str,'("# REWEIGHTING LOWMODE SOLVER, NEV=",I3)') LV%NEV

    tol    = EIGEN_TOL
    iiter  = MAX_ITER
    call get_low_modes(LV,this%target_phys_param,iout_lowmode_log,tol,iiter,u)

    call delete(LV)
    deallocate(LV)

  endif

1000 continue

  allocate(eta)
  call new(eta)

  !------------------
  ! NS step loop
  !------------------
  NS0 = NS_org
  NS1 = NS_org+dNS
  do isteps_ns = 1, this%num_reweighting_steps + 2

    if ( NS0 > NS_target ) then
      NS0 = NS_target
    endif 
    if ( NS1 > NS_target ) then
      NS1 = NS_target
    endif 

    if ( NS0 == NS1 .and. NS1 == NS_target ) then
      if (is_lowprojd) then

        call assign(phys_param_ns0,this%hmc_phys_param)
        call set_NS(phys_param_ns0,NS0)
        call set_coef(phys_param_ns0)

        call assign(phys_param_ns1,this%target_phys_param)
        call set_NS(phys_param_ns1,NS1)
        call set_coef(phys_param_ns1)

      else

        call assign(phys_param_ns0,this%hmc_phys_param)
        call set_NS(phys_param_ns0,NS0)
        call set_coef(phys_param_ns0)

        call assign(phys_param_ns1,this%hmc_phys_param)
        call set_NS(phys_param_ns1,NS1)
        call set_coef(phys_param_ns1)

      endif

    else

      call assign(phys_param_ns0,this%hmc_phys_param)
      call set_NS(phys_param_ns0,NS0)
      call set_coef(phys_param_ns0)

      call assign(phys_param_ns1,this%hmc_phys_param)
      call set_NS(phys_param_ns1,NS1)
      call set_coef(phys_param_ns1)

    endif

#ifdef _DEBUG
    if (0 == nodeid) then
      call print(phys_param_ns0)
      call print(phys_param_ns1)
    endif
#endif

    ave_dS  = 0.0_DP
    ave_exp = 0.0_DP

    !------------------------------
    ! noise loop
    !------------------------------
    do inoise = 1, this%num_noise_rew

      call set_gaussian_noise(eta)

      allocate(v,w)
      call new(v)
      call new(w)


      !------------------------------------
      !  v = Dov(NS0) eta
      !------------------------------------
      write(iout_solver_log,'("# REWEIGHTING SOLVER: v=Dov(NS0)*eta, NOISE=",I4," NS=",I3)')inoise,NS0
#ifdef _DEBUG
      if (0 == nodeid) write(*,'(A)')TRIM(str)
#endif
      iiter = MAX_ITER
      tol = this%reweighting_inv%tol
      call assign_mult_ov(phys_param_ns0,iout_solver_log,tol,iiter,v,eta,u)
      call inc(this%reweighting_inv,iiter)

      !------------------------------------
      ! w = Dov(NS1) \ v
      !------------------------------------
      write(iout_solver_log,'("# REWEIGHTING SOLVER: v=Dov(NS1)\eta, NOISE=",I4," NS=",I3)')inoise,NS1
#ifdef _DEBUG
      if (0 == nodeid) write(*,'(A)')TRIM(str)
#endif
      iiter = MAX_ITER
      tol = this%reweighting_inv%tol

      if ( NS0 == NS1 .and. NS1 == NS_target ) then
        if (is_lowprojd) then
#ifdef _DEBUG
          if (0==nodeid) write(*,'("Low mode projection at NS=",I3)')NS1
#endif
          call assign_inv_mult_ov(phys_param_ns1,LV,iout_solver_log,tol,iiter,w,v,u)
        else
          call assign_inv_mult_ov(phys_param_ns1,   iout_solver_log,tol,iiter,w,v,u)
        endif
      else
        call assign_inv_mult_ov(phys_param_ns1,iout,tol,iiter,w,v,u)
      endif

      call inc(this%reweighting_inv,iiter)

      !-------------
      ! v = w + eta
      ! w = w - eta
      !-------------
      call assign_add(v,w,eta)
      call  accum_sub(w,eta)
    
      !------------------------------------------------------
      ! exp_weight = |Dov(target)\Dov(HMC) eta|^2 - |eta|^2
      !            = Re(w'*v)
      ! weight = exp(-exp_weight)
      !------------------------------------------------------
      rtmp1 = real(prod(w,v),kind=DP)
      rtmp2 = exp(-rtmp1)


      write(iout_reweighting_log,'(I7," dS_=",E24.15," exp(-dS)_=",E24.15," NOISE_=",I3," NS0_=",I3," NS1_=",I3)') &
 &         itraj,rtmp1,rtmp2,inoise,NS0,NS1
#ifdef _DEBUG
      if (0 == nodeid) write(*,'(A)')TRIM(str)
#endif

      ave_dS  = ave_dS  + rtmp1
      ave_exp = ave_exp + rtmp2

      deallocate(v,w)

      !----------------------
      ! check GW relation
      !----------------------
      tol = this%reweighting_inv%tol
      call check_GW_relation_tdwov(phys_param_ns0,                         &
  &                                iout_gw_rel_log=iout_reweighting_log,   &
  &                                iout_solver_log=iout_solver_log, &
  &                                tol=tol,eta=eta,itraj=itraj,u=u)

      tol = this%reweighting_inv%tol
      if ( NS0 == NS1 .and. NS1 == NS_target ) then
        if (is_lowprojd) then
#ifdef _DEBUG
          if (0 == nodeid) write(*,'("Low mode projection at NS=",I3)')NS1
#endif
          call check_GW_relation_tdwov(phys_param_ns1,LV,                      &
  &                                    iout_gw_rel_log=iout_reweighting_log,   &
  &                                    iout_solver_log=iout_solver_log, &
  &                                    tol=tol,eta=eta,itraj=itraj,u=u)
        else
          call check_GW_relation_tdwov(phys_param_ns1,                         &
  &                                    iout_gw_rel_log=iout_reweighting_log,   &
  &                                    iout_solver_log=iout_solver_log, &
  &                                    tol=tol,eta=eta,itraj=itraj,u=u)
        endif
      else
        call check_GW_relation_tdwov(phys_param_ns1,                         &
  &                                  iout_gw_rel_log=iout_reweighting_log,   &
  &                                  iout_solver_log=iout_solver_log, &
  &                                  tol=tol,eta=eta,itraj=itraj,u=u)
      endif


    enddo

    ave_dS  = ave_dS /this%num_noise_rew
    ave_exp = ave_exp/this%num_noise_rew
    write(iout_reweighting_log,'("@",I6," dS_=",E24.15," exp(-dS)_=",E24.15," NOISE_AVERAGED")')itraj,ave_dS,ave_exp

    if ( NS0 == NS1 .and. NS1 == NS_target ) exit

    NS0 = NS0 + dNS
    NS1 = NS1 + dNS

    call delete(phys_param_ns0)
    call delete(phys_param_ns1)

  enddo

  deallocate(eta)

  if (is_lowprojd) then
    call delete(LV)
    deallocate(LV)
  endif


  call toc(etime)
  rtmp0 = get_elapse(etime)
  write(iout_reweighting_log,'("# ETIME_=",E24.15)')rtmp0
  this%run_time = this%run_time + rtmp0


  return
end subroutine

subroutine check_GW_relation_tdwov(this,LV,iout_gw_rel_log,iout_solver_log,tol,eta,itraj,u)
!
! Check chiral symmetry breaking with GW relation
!
!\[
!  0 = |\Delta \eta|^2 /|\eta|^2
!\]
!
! with
!
!\[
! \Delta = \left[ D_{OV}^{-1} \gamma_{5} + gamma_{5} D_{OV}^{-1}
!  - \frac{2}{1+m_q} \left( m_q D_{OV}^{-1} gamma_5 D_{OV}^{-1} + gamma_5 \right) \right]\eta
!\]
!
! and Gaussian noise  $ \eta $ .
!
!
  use hmc_status_class
  implicit none
  type(quark_truncd_dw_overlap), intent(inout) :: this  ! quark action/algorithm parameters
  type(dwf_low_modes), optional, intent(in)    :: LV
  integer,                       intent(in)    :: iout_gw_rel_log
  integer,                       intent(in)    :: iout_solver_log
  real(DP),                      intent(in)    :: tol
  type(field_quark_wg),          intent(inout) :: eta  ! Gaussian noise field
  integer,                       intent(in)    :: itraj
  type(vfield_gluon_wg),         intent(in)    :: u
  type(field_quark_wg), allocatable :: v,w,s,y
  type(quark_truncd_dw_overlap), pointer :: phys_param
  integer  :: i,iiter,NS,iout
  real(DP) :: mq,rtmp0,rtmp1,rtmp2
  character(CHARLEN) :: str
  logical :: is_lowprojd

  if (present(LV)) then
    is_lowprojd = .true.
  else
    is_lowprojd = .false.
  endif

  allocate(v,w,s,y)
  call new(v)
  call new(w)
  call new(s)
  call new(y)

  write(iout_solver_log,'("# GW RELATION SOLVER")')

  NS = get_NS(this)
  mq = get_mass(this)

  !-----------------------
  ! w = gamma5 Dov \ eta
  !-----------------------
  iiter = MAX_ITER
  iout = iout_solver_log
  if (is_lowprojd) then
#ifdef _DEBUG
    if (0 == nodeid) write(*,'("GW_REL with Low mode projection at NS=",I3)')get_NS(this)
#endif
    call assign_inv_mult_ov(this,LV,iout,tol,iiter,w,eta,u)
  else
    call assign_inv_mult_ov(this   ,iout,tol,iiter,w,eta,u)
  endif

  call accum_mult_gamma5(w)

  !----------------------------
  ! y = Dov \ gamma5 Dov \ eta
  !----------------------------
  iiter = MAX_ITER
  iout = iout_solver_log
  if (is_lowprojd) then
#ifdef _DEBUG
    if (0 == nodeid) write(*,'("GW_REL with Low mode projection at NS=",I3)')get_NS(this)
#endif
    call assign_inv_mult_ov(this,LV,iout,tol,iiter,y,w,u)
  else
    call assign_inv_mult_ov(this   ,iout,tol,iiter,y,w,u)
  endif

  !-----------------------
  ! v = gamma_5 eta
  !-----------------------
  call assign_mult_gamma5(v,eta)

  !-----------------------
  ! s = Dov \ gamma_5 eta
  !-----------------------
  iiter = MAX_ITER
  iout = iout_solver_log
  if (is_lowprojd) then
#ifdef _DEBUG
    if (0==nodeid) write(*,'("GW_REL with Low mode projection at NS=",I3)')get_NS(this)
#endif
    call assign_inv_mult_ov(this,LV,iout,tol,iiter,s,v,u)
  else
    call assign_inv_mult_ov(this   ,iout,tol,iiter,s,v,u)
  endif

  !------------
  ! w = w + s
  !------------
  call accum_add(w,s)

  !--------------------
  ! w = w -2mq/(1+m) y
  !--------------------
  rtmp0 = -2.0_DP*mq/(1.0_DP+mq)
  call accum_add_mult(w,y,rtmp0)

  !--------------------
  ! w = w -2/(1+m) v
  !--------------------
  rtmp0 = -2.0_DP/(1.0_DP+mq)
  call accum_add_mult(w,v,rtmp0)

  !---------------------
  ! relative norm
  !---------------------
  rtmp0 = sqrt(abs2(w)/abs2(eta))
  write(iout_gw_rel_log,'("G",I6," GW_REL_=",E24.15," NS_=",I3)')itraj,rtmp0,NS

  deallocate(v,w,s,y)

  return
end subroutine

end module
