#include <config.h>

!
!= SF measurement class
!
!== Version
!
! $Id: sf_measure_class.F90,v 1.5 2011/10/31 07:48:16 ishikawa Exp $
!
module sf_measure_class
!
! This module contains the following SF measurements.
!
!  - SF scheme coupling measurement.
!  - PCAC correlation function measurement with SF boundary. (f_A,f_P,f_A',f_P')
!
  use lattice_class
  use logfile_class
  use quark_prop_class
  use wavefunc_param_class
  use wavefunc_class
  use field_gauge_class
#ifdef _CLOVER
  use quark_clover_class
#else
  use quark_wilson_class
#endif
  use hmc_status_class
  use hmc_stout_class
  implicit none
  private
  public :: sf_measurement
  public :: new, delete
  public :: measure

  integer, parameter :: NSPACE=NX*NY*NZ

  type sf_measurement
  !
  ! SF measurement data (Wilson/Clover)
  !
    character(len=CHARLEN) :: fname_quark  = ""
    character(len=CHARLEN) :: fname_gsf    = "Sf_coupling."
    character(len=CHARLEN) :: fname_pcac   = "Sf_PCAC."
    character(len=CHARLEN) :: fname_solver = "Sf_PCAC_solver_log."
    type(logfile), pointer :: logfile_gsf    => NULL()
    type(logfile), pointer :: logfile_pcac   => NULL()
    type(logfile), pointer :: logfile_solver => NULL()
    type(hmc_status), pointer :: status => NULL()

    real(DP) :: kappa = 0.0_DP
    real(DP) ::   csw = 0.0_DP
    integer  :: nflavor
    real(DP) :: tolerance
    real(DP) :: stout_alpha = 0.0_DP
    integer  :: stout_nstep = 0

    complex(DP) :: fA0(2:NTT), fP0(2:NTT)
    complex(DP) :: fA1(2:NTT), fP1(2:NTT)
    real(DP) :: gSF2inv
    real(DP) :: vSF

  end type

  interface new
    module procedure new_sf_measurement
  end interface

  interface delete
    module procedure delete_sf_measurement
  end interface

  interface measure
    module procedure measure_sf
  end interface

contains

subroutine new_sf_measurement(this,quark_fname,status)
!
! Initialize SF measurement
!
!        this : SF measurement
! quark_fname : SF PCAC quark mass parameter file name
!      status : HMC job status
!
  use comlib, only : comlib_bcast
  use tokenize
  use file_tools_class
  implicit none
  type(sf_measurement),     intent(inout) :: this
  character(len=*),         intent(in)    :: quark_fname
  type(hmc_status), target, intent(in)    :: status
  character(len=CHARLEN) :: str, part
  type(tokenizer) :: token
  integer :: iout,itraj,token_len

  call set_tokenizer(token,' ',' ',' ')

  this%fname_quark = TRIM(quark_fname)
  this%status => status

  if (0==nodeid) then
    iout = search_free_file_unit()
    open(iout,file=TRIM(this%fname_quark),status='old',form='formatted')
    read(iout,*)str
#ifdef _CLOVER
    read(iout,*)this%kappa,this%csw
#else
    this%csw = ZERO
    read(iout,*)this%kappa
#endif
    read(iout,*)this%nflavor
    read(iout,*)this%stout_alpha,this%stout_nstep
    read(iout,*)str
    read(iout,*)this%tolerance
    close(iout)
  endif
#ifndef _singlePU  
  call comlib_bcast(this%kappa,0)
  call comlib_bcast(this%csw,0)
  call comlib_bcast(this%nflavor,0)
  call comlib_bcast(this%tolerance,0)
  call comlib_bcast(this%stout_alpha,0)
  call comlib_bcast(this%stout_nstep,0)
#endif

  allocate(this%logfile_gsf)
  allocate(this%logfile_pcac)
  allocate(this%logfile_solver)

  !===============================================
  ! open SF scheme coupling log file
  !===============================================
  itraj = get_trajectory_number(this%status)
  write(str,'(A,I6.6)')TRIM(ADJUSTL(this%fname_gsf)),itraj
  call new(this%logfile_gsf,TRIM(str))

  call print(this%logfile_gsf,"# SF scheme coupling")
  write(str,'("# traj",8X,"1/(g_SF^2)",16X,"v-term",16X,"k/(g_SF^2)",15X,"k*v-term",20X,"k")')
  call print(this%logfile_gsf,TRIM(str))


  !===============================================
  ! open SF PCAC log file
  !===============================================
  write(str,'(A,I6.6)')TRIM(ADJUSTL(this%fname_pcac)),itraj
  call new(this%logfile_pcac,TRIM(str))

  call print(this%logfile_pcac,"# SF PCAC correlation functions")
  write(str,'("# traj itt",9X,"fA(itt)",17X,"fP(itt)",17X,"fA''(itt)",16X,"fP''(itt)")')
  call print(this%logfile_pcac,TRIM(str))


  !===============================================
  ! open SF PCAC quark solver log file
  !===============================================
  write(str,'(A,I6.6)')TRIM(ADJUSTL(this%fname_solver)),itraj
  call new(this%logfile_solver,TRIM(str))
  call print(this%logfile_solver,"# SF PCAC correlation solver log")


  if (0==nodeid) then
    write(*,'(80("="))')
    write(*,'("  SF measuement : SF coupling and PCAC")')
    write(*,'("        Trajectory :",I10)')  itraj
    if (0 < this%stout_nstep) then
      write(*,'(" Stout semared fermion:")')
      write(*,'("             alpha :",F16.9)')this%stout_alpha
      write(*,'("             nstep :",I3   )')this%stout_nstep
    endif
    write(*,'("       PCAC: Kappa :",F16.9)')this%kappa
#ifdef _CLOVER
    write(*,'("       PCAC:   Csw :",F16.9)')this%csw
#endif
    write(*,'("   SF coupling: Nf :",I3)')   this%nflavor
    write(*,'("  Solver tolerance :",E24.15)')this%tolerance
    write(*,'(80("-"))')
  endif

  return
end subroutine

subroutine delete_sf_measurement(this)
!
! Delete SF measurement
!
  implicit none
  type(sf_measurement),      intent(inout) :: this
  this%status => NULL()
  call delete(this%logfile_gsf)
  call delete(this%logfile_pcac)
  call delete(this%logfile_solver)
  deallocate(this%logfile_gsf)
  deallocate(this%logfile_pcac)
  deallocate(this%logfile_solver)
  this%logfile_gsf => NULL()
  this%logfile_pcac => NULL()
  this%logfile_solver => NULL()
  if (0==nodeid) then
    write(*,'(80("="))')
  endif
  return
end subroutine

subroutine measure_sf(this,gauge,u)
!
! Measure SF scheme coupling and PCAC correlation functions
!
!   this : SF measurement
!  gauge : gauge action couplings (SF boundary couplings are used)
!      u : gauge link
!
  use hmc_gluon_class, only : gluon_rg_sf
  implicit none
  type(sf_measurement),  intent(inout) :: this
  type(gluon_rg_sf),     intent(in) :: gauge
  type(vfield_gluon_wg), intent(in) :: u

  type(quark_prop), allocatable :: qp
  type(qprop_params), target :: params
  type(src_wavefunc_obj), target, allocatable :: wf
  type(stout_smearing_parameters) :: stout


  !========================================
  ! compute SF scheme couplings
  !========================================
  call get_sf_scheme_coupling(this,gauge,u)

  !========================================
  ! compute SF PCAC corelation functions
  !========================================
  call new(params)
  call new(params%mass,id=0)
  call set_kappa(params%mass,this%kappa)
#ifdef _CLOVER
  call set_csw(params%mass,this%csw)
#endif
  params%tol = this%tolerance
  params%solver_log => this%logfile_solver
!  params%smear_sous = 1
!  params%smear_sink = 0

  allocate(qp,wf)
  call new(qp)
  call new(wf)
  qp%parameter     => params
  qp%wavefunc_sous => wf
  qp%smear_sous    = 1
  qp%wavefunc_sink => NULL()
  qp%smear_sink    = 0

  if (0 < this%stout_nstep) then
    call new(stout,this%stout_alpha,this%stout_nstep)
    call smear(stout,u)
    call set_sf_source(qp,stout%usmr(this%stout_nstep))
    call get_prop(qp,stout%usmr(this%stout_nstep))
    call delete(stout)
  else
    call set_sf_source(qp,u)
    call get_prop(qp,u)
  endif
  call get_pcac(this,qp)

  call delete(qp)
  call delete(wf)
  call delete(params)
  deallocate(qp,wf)

  return
end subroutine

subroutine set_sf_source(qp,u)
!
! set SF PCAC correlation function boundary source fields on a quark propagator
!
! itt=2    qp(ic,jc,is=js,js=1,2,t,z,y,x) =  (U_4(t-1,z,y,x)^+)_ic,jc0
! itt=NTT  qp(ic,jc,is=js,js=3,4,t,z,y,x) =  (U_4(t,z,y,x))_ic,jc0
!
!    qp : quark propagator
!     u : gauge link field
!
  use error_class
  implicit none
  type(quark_prop),      intent(inout) :: qp
  type(vfield_gluon_wg), intent(in)    :: u
  integer :: ieo,ix,iy,iz,ieoxyz,ith,ic,jc,is,js,isp,itt

  if (allocated(qp%prop)) deallocate(qp%prop)
  allocate(qp%prop(COL,COL,SPIN,SPIN,NZ*NY*NX,NTT))

!$OMP PARALLEL DO PRIVATE(itt,isp,js,is,jc,ic)
  do itt=1,NTT
  do isp=1,NSPACE
    do js=1,SPIN
    do is=1,SPIN
    do jc=1,COL
    do ic=1,COL
      qp%prop(ic,jc,is,js,isp,itt) = Z0
    enddo
    enddo
    enddo
    enddo
  enddo
  enddo

  do ieo=0,1
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,isp,itt,js,is,jc,ic)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+iz+iy+ix+ieo,2)
    isp = ispace(ix,iy,iz)

!=======================================
! Set itt = 2 boundary
!
!     ith = 1-ieoxyz,..,NTH-ieoxyz
!     itt = 2*ith + ieoxyz
!     itt = 2 => ith = 1, ieoxyz = 0 
!     itt = 1 => ith = 0, ieoxyz = 1
!
!  Dirac rep.  (1+gamma_4)/2 projection
! set spin = 1,2 (upper 2 spinor) only.
!
!=======================================
    if (1==ieoxyz) then
      itt = 2
      do is=1,2    !<== upper 2 spinor
      do jc=1,COL
      do ic=1,COL
        qp%prop(ic,jc,is,is,isp,itt) = conjg(u%eo(ieo)%mu(4)%s(0,iz,iy,ix)%u(jc,ic))  ! <== (U_4(t=1))^+_(ic,jc)
      enddo
      enddo
      enddo
    endif

!=======================================
! Set itt = NTT boundary
!
!     ith = 1-ieoxyz,..,NTH-ieoxyz
!     itt = 2*ith + ieoxyz
!     itt = NTT => ith = NTH, ieoxyz = 0 
!
!  Dirac rep.  (1-gamma_4)/2 projection
! set spin = 3,4 (lower 2 spinor) only.
!
!=======================================
    if (0==ieoxyz) then
      itt = NTT
      do is=3,4    !<== lower 2 spinor
      do jc=1,COL
      do ic=1,COL
        qp%prop(ic,jc,is,is,isp,itt) = u%eo(ieo)%mu(4)%s(NTH,iz,iy,ix)%u(ic,jc)       ! <== U_4(t=NTT)_(ic,jc)
      enddo
      enddo
      enddo
    endif

  enddo
  enddo
  enddo
  enddo

  return
end subroutine
  
subroutine get_pcac(this,qp)
!
! measure SF PCAC correlation functions.
!
!   fA(x_0)   = 1/2/Vol Sum_{vec{x}} Tr[qp^+(vec{x},x_0) [-gamma_0] qp(vec{x},x_0)]
!   fP(x_0)   = 1/2/Vol Sum_{vec{x}} Tr[qp^+(vec{x},x_0)            qp(vec{x},x_0)]
!   use qp(:,:,:,js=1,2) only (Dirac Rep.(1+gamma4)/2)
! 
!   fA'(T-x_0)= 1/2/Vol Sum_{vec{x}} Tr[qp^+(vec{x},x_0) [+gamma_0] qp(vec{x},x_0)]
!   fP'(T-x_0)= 1/2/Vol Sum_{vec{x}} Tr[qp^+(vec{x},x_0)            qp(vec{x},x_0)]
!   use qp(:,:,:,js=3,4) only (Dirac Rep.(1-gamma4)/2)
! 
!     this : sf measuremente 
!       qp : quark propagator
! 
  use comlib, only : comlib_sumcast
  use field_fermion_class, only : igamma, zgamma
  implicit none
  type(sf_measurement), intent(inout) :: this
  type(quark_prop),     intent(in) :: qp
  integer :: itt,isp,ic,is,jc,js
  integer :: isnk
  complex(DP) :: zgm
  complex(DP) :: ztmp0_A,ztmp1_A
  complex(DP) :: ztmp0_P,ztmp1_P
  integer :: itraj
  character(len=CHARLEN) :: str

  this%fA0(:)= Z0
  this%fP0(:)= Z0
  this%fA1(:)= Z0
  this%fP1(:)= Z0

  do itt=2,NTT
  ztmp0_A = Z0; ztmp0_P = Z0
  ztmp1_A = Z0; ztmp1_P = Z0
!$OMP PARALLEL DO PRIVATE(isp,js,is,jc,ic,isnk,zgm) REDUCTION(+:ztmp0_A,ztmp0_P,ztmp1_A,ztmp1_P)
  do isp = 1,NSPACE
!********* fA
    do js=1,SPIN/2
    do is=1,SPIN
      isnk = igamma(is,4)
      zgm  = zgamma(is,4)
      do jc=1,COL
      do ic=1,COL
        ztmp0_A = ztmp0_A + zgm*qp%prop(ic,jc,isnk,js  ,isp,itt)   &
 &                       *conjg(qp%prop(ic,jc,is  ,js  ,isp,itt))
        ztmp1_A = ztmp1_A + zgm*qp%prop(ic,jc,isnk,js+2,isp,itt)   &
 &                       *conjg(qp%prop(ic,jc,is  ,js+2,isp,itt))
      enddo
      enddo
    enddo
    enddo
!********* fP
    do js=1,SPIN/2
    do is=1,SPIN
      do jc=1,COL
      do ic=1,COL
        ztmp0_P = ztmp0_P +     qp%prop(ic,jc,is,js  ,isp,itt)  &
     &                   *conjg(qp%prop(ic,jc,is,js  ,isp,itt))
        ztmp1_P = ztmp1_P +     qp%prop(ic,jc,is,js+2,isp,itt)  &
     &                   *conjg(qp%prop(ic,jc,is,js+2,isp,itt))
      enddo
      enddo
    enddo
    enddo
  enddo
  this%fA0(itt)       = ztmp0_A
  this%fP0(itt)       = ztmp0_P
  this%fA1(NTT+2-itt) = ztmp1_A
  this%fP1(NTT+2-itt) = ztmp1_P
  enddo

#ifndef _singlePU
  call comlib_sumcast(this%fA0(:))
  call comlib_sumcast(this%fP0(:))
  call comlib_sumcast(this%fA1(:))
  call comlib_sumcast(this%fP1(:))
#endif

  do itt=2,NTT
    this%fA0(itt)=-this%fA0(itt)/(2*NTX*NTY*NTZ)
    this%fP0(itt)= this%fP0(itt)/(2*NTX*NTY*NTZ)
    this%fA1(itt)= this%fA1(itt)/(2*NTX*NTY*NTZ)
    this%fP1(itt)= this%fP1(itt)/(2*NTX*NTY*NTZ)
  enddo

  itraj = get_trajectory_number(this%status)

  if (0==nodeid) then
    do itt=2,NTT
      write(str,'(I6,I4,4E24.16)')itraj,itt, &
 &    REAL(this%fA0(itt),kind=DP),REAL(this%fP0(itt),kind=DP), &
 &    REAL(this%fA1(itt),kind=DP),REAL(this%fP1(itt),kind=DP)
      call print(this%logfile_pcac,TRIM(str))
    enddo
  endif

  return
end subroutine

subroutine get_sf_scheme_coupling(this,gauge,u)
!
! Calculate SF scheme coupling
!
!  d Gamma|                  1     
!  -------|         = k ( -------- - mu v )
!  d  eta |_eta=0         (g_SF)^2 
!
!
!  this : SF parameters
! gauge : gauge action parameters
!     u : 
!
  use comlib, only : comlib_sumcast
  use hmc_gluon_class, only : gluon_rg_sf
  implicit none
  type(sf_measurement), intent(inout) :: this
  type(gluon_rg_sf),       intent(in) :: gauge
  type(vfield_gluon_wg),   intent(in) :: u

  integer :: itraj
  character(len=CHARLEN) :: str

  real(DP)    :: dSGg,dSGv
  complex(DP) :: w23(COL,COL),w41(COL,COL)
  complex(DP) :: w25(COL,COL),w61(COL,COL)
  complex(DP) :: w2341(COL,COL),w2561(COL,COL)
  complex(DP) :: uk0(COL,COL),vk0(COL,COL)
  complex(DP), parameter :: zi=(0.0_DP,1.0_DP)
  real(DP) :: omega(COL,NDIM-1),vi(COL,NDIM-1)
  real(DP) :: knorm,th(NDIM-1)
  real(DP) :: trwUk0,trwVk0
  integer :: ieo,ioe,ix,iy,iz,ieoxyz,ic,jc,mu
  integer :: ix2,iy2,iz2
#ifdef _RGGAUGE
  real(DP) :: dSGg1,dSGv1,dSGg2,dSGv2
  complex(DP) :: we4(COL,COL),we41(COL,COL),w23d(COL,COL)
  complex(DP) :: wi6(COL,COL),wi61(COL,COL),w25h(COL,COL)
  complex(DP) :: w23de41(COL,COL),w25hi61(COL,COL)
  complex(DP) :: w78(COL,COL),w478(COL,COL),w123(COL,COL)
  complex(DP) :: w98(COL,COL),w698(COL,COL),w125(COL,COL)
  complex(DP) :: w123478(COL,COL),w125698(COL,COL)
  integer :: ix8,iy8,iz8
#endif

  omega(1,1)= 1.0_DP/NTX
  omega(1,2)= 1.0_DP/NTY
  omega(1,3)= 1.0_DP/NTZ
  omega(2,1)=-0.5_DP/NTX
  omega(2,2)=-0.5_DP/NTY
  omega(2,3)=-0.5_DP/NTZ
  omega(3,1)=-0.5_DP/NTX
  omega(3,2)=-0.5_DP/NTY
  omega(3,3)=-0.5_DP/NTZ

  vi(1,1)= 0.0_DP
  vi(1,2)= 0.0_DP
  vi(1,3)= 0.0_DP
  vi(2,1)= 1.0_DP/NTX
  vi(2,2)= 1.0_DP/NTY
  vi(2,3)= 1.0_DP/NTZ
  vi(3,1)=-1.0_DP/NTX
  vi(3,2)=-1.0_DP/NTY
  vi(3,3)=-1.0_DP/NTZ

  th(1) = PI/COL/NTT/NTX
  th(2) = PI/COL/NTT/NTY
  th(3) = PI/COL/NTT/NTZ

!======================================================
! Normalizetion from action on a classical SF field
!======================================================
#ifdef _RGGAUGE
  knorm=4*NTX*NTY*NTZ*( gauge%cRG(0)*( (sin(2*th(1))+sin(th(1)))/NTX     &
 &                                    +(sin(2*th(2))+sin(th(2)))/NTY     &
 &                                    +(sin(2*th(3))+sin(th(3)))/NTZ)    &
 &                   +4*gauge%cRG(1)*( (sin(4*th(1))+sin(2*th(1)))/NTX   &
 &                                    +(sin(4*th(2))+sin(2*th(2)))/NTY   &
 &                                    +(sin(4*th(3))+sin(2*th(3)))/NTZ) )
#else
  knorm=4*NTX*NTY*NTZ*( (sin(2*th(1))+sin(th(1)))/NTX    &
 &                     +(sin(2*th(2))+sin(th(2)))/NTY    &
 &                     +(sin(2*th(3))+sin(th(3)))/NTZ )
#endif

  dSGg=0.0_DP  ! for 1/gSF^2
  dSGv=0.0_DP  ! for v
#ifdef _RGGAUGE
  dSGg1=0.0_DP  ! for 1/gSF^2 (rectangular part 1x2)
  dSGv1=0.0_DP  ! for v       (rectangular part 1x2)
  dSGg2=0.0_DP  ! for 1/gSF^2 (rectangular part 2x1)
  dSGv2=0.0_DP  ! for v       (rectangular part 2x1)
#endif

#ifdef _RGGAUGE
!$OMP PARALLEL DO &
!$OMP& PRIVATE(ix,iy,iz,ieoxyz,ieo,ioe,mu,ic,jc, &
!$OMP&         ix2,iy2,iz2,w23,w25,w41,w61,w2341,w2561,uk0,vk0,trwUk0,trwVk0, &
!$OMP&         ix8,iy8,iz8,we4,wi6,we41,w23d,wi61,w25h,w23de41,w25hi61, &
!$OMP&                     w78,w98,w478,w123,w698,w125,w123478,w125698) &
!$OMP& REDUCTION(+:dSGg,dSGv,dSGg1,dSGv1,dSGg2,dSGv2)
#else
!$OMP PARALLEL DO &
!$OMP& PRIVATE(ix,iy,iz,ieoxyz,ieo,ioe,mu,ic,jc, &
!$OMP&         ix2,iy2,iz2,w23,w25,w41,w61,w2341,w2561,uk0,vk0,trwUk0,trwVk0) &
!$OMP& REDUCTION(+:dSGg,dSGv)
#endif
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz,2)

    do mu=1,NDIM-1  !<=== Spatial only
      ix2=ix+isx(mu) ! n + e_k
      iy2=iy+isy(mu)
      iz2=iz+isz(mu)
#ifdef _RGGAUGE
      ix8=ix-isx(mu) ! n - e_k
      iy8=iy-isy(mu)
      iz8=iz-isz(mu)
#endif
      if (1 == ieoxyz) then
        ieo = 0
        ioe = 1-ieo
      else
        ieo = 1
        ioe = 1-ieo
      endif

!******************************************
!*     v                  t=NTT+1     v
!*   n+4                  n         n+k
!*    4---<---3            1--->---2
!*    |       |            |       |
!*    v       ^            ^       v
!*    |       |            |       |
!*    1--->---2  v         6---<---5
!*   n         n+k       n-4
!*   t = 1                          for plaquette
!******************************************
      do jc=1,COL
      do ic=1,COL
        w23(ic,jc)=       u%eo(ieo)%mu(mu)%s(  0,iz ,iy ,ix )%u(ic, 1)*   &
 &                        u%eo(ioe)%mu( 4)%s(  0,iz2,iy2,ix2)%u( 1,jc)    &
 &                       +u%eo(ieo)%mu(mu)%s(  0,iz ,iy ,ix )%u(ic, 2)*   &
 &                        u%eo(ioe)%mu( 4)%s(  0,iz2,iy2,ix2)%u( 2,jc)    &
 &                       +u%eo(ieo)%mu(mu)%s(  0,iz ,iy ,ix )%u(ic, 3)*   &
 &                        u%eo(ioe)%mu( 4)%s(  0,iz2,iy2,ix2)%u( 3,jc)
        w25(ic,jc)=       u%eo(ieo)%mu(mu)%s(NTH,iz ,iy ,ix )%u(ic, 1)*   &
 &                  conjg(u%eo(ieo)%mu( 4)%s(NTH,iz2,iy2,ix2)%u(jc, 1))   &
 &                 +      u%eo(ieo)%mu(mu)%s(NTH,iz ,iy ,ix )%u(ic, 2)*   &
 &                  conjg(u%eo(ieo)%mu( 4)%s(NTH,iz2,iy2,ix2)%u(jc, 2))   &
 &                 +      u%eo(ieo)%mu(mu)%s(NTH,iz ,iy ,ix )%u(ic, 3)*   &
 &                  conjg(u%eo(ieo)%mu( 4)%s(NTH,iz2,iy2,ix2)%u(jc, 3))
      enddo
      enddo

      do jc=1,COL
      do ic=1,COL
        w41(ic,jc)= conjg(u%eo(ioe)%mu(mu)%s(  1,iz,iy,ix)%u( 1,ic))*   &
 &                  conjg(u%eo(ieo)%mu( 4)%s(  0,iz,iy,ix)%u(jc, 1))    &
 &                 +conjg(u%eo(ioe)%mu(mu)%s(  1,iz,iy,ix)%u( 2,ic))*   &
 &                  conjg(u%eo(ieo)%mu( 4)%s(  0,iz,iy,ix)%u(jc, 2))    &
 &                 +conjg(u%eo(ioe)%mu(mu)%s(  1,iz,iy,ix)%u( 3,ic))*   &
 &                  conjg(u%eo(ieo)%mu( 4)%s(  0,iz,iy,ix)%u(jc, 3))
        w61(ic,jc)= conjg(u%eo(ioe)%mu(mu)%s(NTH,iz,iy,ix)%u( 1,ic))*   &
 &                        u%eo(ioe)%mu( 4)%s(NTH,iz,iy,ix)%u( 1,jc)     &
 &                 +conjg(u%eo(ioe)%mu(mu)%s(NTH,iz,iy,ix)%u( 2,ic))*   &
 &                        u%eo(ioe)%mu( 4)%s(NTH,iz,iy,ix)%u( 2,jc)     &
 &                 +conjg(u%eo(ioe)%mu(mu)%s(NTH,iz,iy,ix)%u( 3,ic))*   &
 &                        u%eo(ioe)%mu( 4)%s(NTH,iz,iy,ix)%u( 3,jc)
      enddo
      enddo

      do jc=1,COL
      do ic=1,COL
        w2341(ic,jc)=w23(ic,1)*w41(1,jc)  &
 &                  +w23(ic,2)*w41(2,jc)  &
 &                  +w23(ic,3)*w41(3,jc)
        w2561(ic,jc)=w25(ic,1)*w61(1,jc)  &
 &                  +w25(ic,2)*w61(2,jc)  &
 &                  +w25(ic,3)*w61(3,jc)
      enddo
      enddo
      do jc=1,COL
      do ic=1,COL
        uk0(ic,jc)=(w2341(ic,jc)-conjg(w2341(jc,ic)))  &
 &                -(w2561(ic,jc)-conjg(w2561(jc,ic)))
        vk0(ic,jc)=(w2341(ic,jc)-conjg(w2341(jc,ic)))  &
 &                +(w2561(ic,jc)-conjg(w2561(jc,ic)))
      enddo
      enddo
      trwUk0 = -REAL(ZI*( omega(1,mu)*uk0(1,1)+omega(2,mu)*uk0(2,2)+omega(3,mu)*uk0(3,3)),kind=DP)
      trwVk0 = -REAL(ZI*(    vi(1,mu)*vk0(1,1)+   vi(2,mu)*vk0(2,2)+   vi(3,mu)*vk0(3,3)),kind=DP)
      dSGg=dSGg+trwUk0
      dSGv=dSGv+trwVk0


#ifdef _RGGAUGE
!******************************************
!*
!*     v                  t=NTT+1     v
!* n+2*4                  n         n+k
!*    e---<---d            1--->---2
!*    |       |            |       |
!*    v       ^            ^       v
!*    |       |            |       |
!*    4       3            6       5
!*    |       |            |       |
!*    v       ^            ^       v
!*    |       |            |       |
!*    1--->---2  v         i---<---h
!*   n         n+k       n-2*4
!*   t = 1                          for rectangle (RG)
!*               v           v        t=NTT+1    v
!*             n+4         n-k        n        n+k
!*     7---<---4---<---3      8--->---1--->---2
!*     |               |      |               |
!*     v               ^      ^               v
!*     |               |      |               |
!*   v 8--->---1--->---2  v   9---<---6---<---5
!* n-k         n        n+k          n-4
!*             t = 1                for rectangle (RG)
!*
!******************************************
      do jc=1,COL
      do ic=1,COL
        we4(ic,jc)= conjg(u%eo(ieo)%mu(mu)%s(    1,iz,iy,ix)%u( 1,ic))*   &
 &                  conjg(u%eo(ioe)%mu( 4)%s(    1,iz,iy,ix)%u(jc, 1))    &
 &                 +conjg(u%eo(ieo)%mu(mu)%s(    1,iz,iy,ix)%u( 2,ic))*   &
 &                  conjg(u%eo(ioe)%mu( 4)%s(    1,iz,iy,ix)%u(jc, 2))    &
 &                 +conjg(u%eo(ieo)%mu(mu)%s(    1,iz,iy,ix)%u( 3,ic))*   &
 &                  conjg(u%eo(ioe)%mu( 4)%s(    1,iz,iy,ix)%u(jc, 3))
        wi6(ic,jc)= conjg(u%eo(ieo)%mu(mu)%s(NTH-1,iz,iy,ix)%u( 1,ic))*   &
 &                        u%eo(ieo)%mu( 4)%s(NTH-1,iz,iy,ix)%u( 1,jc)     &
 &                 +conjg(u%eo(ieo)%mu(mu)%s(NTH-1,iz,iy,ix)%u( 2,ic))*   &
 &                        u%eo(ieo)%mu( 4)%s(NTH-1,iz,iy,ix)%u( 2,jc)     &
 &                 +conjg(u%eo(ieo)%mu(mu)%s(NTH-1,iz,iy,ix)%u( 3,ic))*   &
 &                        u%eo(ieo)%mu( 4)%s(NTH-1,iz,iy,ix)%u( 3,jc)
      enddo
      enddo

      do jc=1,COL
      do ic=1,COL
        we41(ic,jc)=we4(ic,1)*conjg(u%eo(ieo)%mu( 4)%s(    0,iz ,iy ,ix )%u(jc, 1))  &
 &                 +we4(ic,2)*conjg(u%eo(ieo)%mu( 4)%s(    0,iz ,iy ,ix )%u(jc, 2))  &
 &                 +we4(ic,3)*conjg(u%eo(ieo)%mu( 4)%s(    0,iz ,iy ,ix )%u(jc, 3))
        w23d(ic,jc)=w23(ic,1)*      u%eo(ieo)%mu( 4)%s(    1,iz2,iy2,ix2)%u( 1,jc)   &
 &                 +w23(ic,2)*      u%eo(ieo)%mu( 4)%s(    1,iz2,iy2,ix2)%u( 2,jc)   &
 &                 +w23(ic,3)*      u%eo(ieo)%mu( 4)%s(    1,iz2,iy2,ix2)%u( 3,jc)
        wi61(ic,jc)=wi6(ic,1)*      u%eo(ioe)%mu( 4)%s(NTH  ,iz ,iy ,ix )%u( 1,jc)   &
 &                 +wi6(ic,2)*      u%eo(ioe)%mu( 4)%s(NTH  ,iz ,iy ,ix )%u( 2,jc)   &
 &                 +wi6(ic,3)*      u%eo(ioe)%mu( 4)%s(NTH  ,iz ,iy ,ix )%u( 3,jc)
        w25h(ic,jc)=w25(ic,1)*conjg(u%eo(ioe)%mu( 4)%s(NTH-1,iz2,iy2,ix2)%u(jc, 1))  &
 &                 +w25(ic,2)*conjg(u%eo(ioe)%mu( 4)%s(NTH-1,iz2,iy2,ix2)%u(jc, 2))  &
 &                 +w25(ic,3)*conjg(u%eo(ioe)%mu( 4)%s(NTH-1,iz2,iy2,ix2)%u(jc, 3))
      enddo
      enddo

      do jc=1,COL
      do ic=1,COL
        w23de41(ic,jc)=w23d(ic,1)*we41(1,jc)  &
 &                    +w23d(ic,2)*we41(2,jc)  &
 &                    +w23d(ic,3)*we41(3,jc)
        w25hi61(ic,jc)=w25h(ic,1)*wi61(1,jc)  &
 &                    +w25h(ic,2)*wi61(2,jc)  &
 &                    +w25h(ic,3)*wi61(3,jc)
      enddo
      enddo
      do jc=1,COL
      do ic=1,COL
        uk0(ic,jc)=(w23de41(ic,jc)-conjg(w23de41(jc,ic)))  &
 &                -(w25hi61(ic,jc)-conjg(w25hi61(jc,ic)))
        vk0(ic,jc)=(w23de41(ic,jc)-conjg(w23de41(jc,ic)))  &
 &                +(w25hi61(ic,jc)-conjg(w25hi61(jc,ic)))
      enddo
      enddo
      trwUk0 = -REAL(ZI*( omega(1,mu)*uk0(1,1)+omega(2,mu)*uk0(2,2)+omega(3,mu)*uk0(3,3)),kind=DP)
      trwVk0 = -REAL(ZI*(    vi(1,mu)*vk0(1,1)+   vi(2,mu)*vk0(2,2)+   vi(3,mu)*vk0(3,3)),kind=DP)
      dSGg1=dSGg1+trwUk0
      dSGv1=dSGv1+trwVk0


      do jc=1,COL
      do ic=1,COL
        w78(ic,jc)= conjg(u%eo(ieo)%mu(mu)%s(  1,iz8,iy8,ix8)%u( 1,ic))*   &
 &                  conjg(u%eo(ioe)%mu( 4)%s(  0,iz8,iy8,ix8)%u(jc, 1))    &
 &                 +conjg(u%eo(ieo)%mu(mu)%s(  1,iz8,iy8,ix8)%u( 2,ic))*   &
 &                  conjg(u%eo(ioe)%mu( 4)%s(  0,iz8,iy8,ix8)%u(jc, 2))    &
 &                 +conjg(u%eo(ieo)%mu(mu)%s(  1,iz8,iy8,ix8)%u( 3,ic))*   &
 &                  conjg(u%eo(ioe)%mu( 4)%s(  0,iz8,iy8,ix8)%u(jc, 3))
        w98(ic,jc)= conjg(u%eo(ieo)%mu(mu)%s(NTH,iz8,iy8,ix8)%u( 1,ic))*   &
 &                        u%eo(ieo)%mu( 4)%s(NTH,iz8,iy8,ix8)%u( 1,jc)     &
 &                 +conjg(u%eo(ieo)%mu(mu)%s(NTH,iz8,iy8,ix8)%u( 2,ic))*   &
 &                        u%eo(ieo)%mu( 4)%s(NTH,iz8,iy8,ix8)%u( 2,jc)     &
 &                 +conjg(u%eo(ieo)%mu(mu)%s(NTH,iz8,iy8,ix8)%u( 3,ic))*   &
 &                        u%eo(ieo)%mu( 4)%s(NTH,iz8,iy8,ix8)%u( 3,jc)
      enddo
      enddo

      do jc=1,COL
      do ic=1,COL
        w478(ic,jc)=conjg(u%eo(ioe)%mu(mu)%s(  1,iz ,iy ,ix )%u( 1,ic))*w78(1,jc)   &
 &                 +conjg(u%eo(ioe)%mu(mu)%s(  1,iz ,iy ,ix )%u( 2,ic))*w78(2,jc)   &
 &                 +conjg(u%eo(ioe)%mu(mu)%s(  1,iz ,iy ,ix )%u( 3,ic))*w78(3,jc)
        w123(ic,jc)=      u%eo(ioe)%mu(mu)%s(  0,iz8,iy8,ix8)%u(ic, 1) *w23(1,jc)   &
 &                 +      u%eo(ioe)%mu(mu)%s(  0,iz8,iy8,ix8)%u(ic, 2) *w23(2,jc)   &
 &                 +      u%eo(ioe)%mu(mu)%s(  0,iz8,iy8,ix8)%u(ic, 3) *w23(3,jc)
        w698(ic,jc)=conjg(u%eo(ioe)%mu(mu)%s(NTH,iz ,iy ,ix )%u( 1,ic))*w98(1,jc)   &
 &                 +conjg(u%eo(ioe)%mu(mu)%s(NTH,iz ,iy ,ix )%u( 2,ic))*w98(2,jc)   &
 &                 +conjg(u%eo(ioe)%mu(mu)%s(NTH,iz ,iy ,ix )%u( 3,ic))*w98(3,jc)
        w125(ic,jc)=      u%eo(ioe)%mu(mu)%s(NTH,iz8,iy8,ix8)%u(ic, 1) *w25(1,jc)   &
 &                 +      u%eo(ioe)%mu(mu)%s(NTH,iz8,iy8,ix8)%u(ic, 2) *w25(2,jc)   &
 &                 +      u%eo(ioe)%mu(mu)%s(NTH,iz8,iy8,ix8)%u(ic, 3) *w25(3,jc)
      enddo
      enddo

      do jc=1,COL
      do ic=1,COL
        w123478(ic,jc)=w123(ic,1)*w478(1,jc)  &
 &                    +w123(ic,2)*w478(2,jc)  &
 &                    +w123(ic,3)*w478(3,jc)
        w125698(ic,jc)=w125(ic,1)*w698(1,jc)  &
 &                    +w125(ic,2)*w698(2,jc)  &
 &                    +w125(ic,3)*w698(3,jc)
      enddo
      enddo
      do jc=1,COL
      do ic=1,COL
        uk0(ic,jc)=(w123478(ic,jc)-conjg(w123478(jc,ic)))  &
 &                -(w125698(ic,jc)-conjg(w125698(jc,ic)))
        vk0(ic,jc)=(w123478(ic,jc)-conjg(w123478(jc,ic)))  &
 &                +(w125698(ic,jc)-conjg(w125698(jc,ic)))
      enddo
      enddo
      trwUk0 = -REAL(ZI*( omega(1,mu)*uk0(1,1)+omega(2,mu)*uk0(2,2)+omega(3,mu)*uk0(3,3)),kind=DP)
      trwVk0 = -REAL(ZI*(    vi(1,mu)*vk0(1,1)+   vi(2,mu)*vk0(2,2)+   vi(3,mu)*vk0(3,3)),kind=DP)
      dSGg2=dSGg2+trwUk0
      dSGv2=dSGv2+trwVk0
#endif

    enddo ! end of do mu
  enddo
  enddo
  enddo  ! end of do ix,iy,iz

#ifndef _singlePU
  call comlib_sumcast(dSGg)
  call comlib_sumcast(dSGv)
#ifdef _RGGAUGE
  call comlib_sumcast(dSGg1)
  call comlib_sumcast(dSGv1)
  call comlib_sumcast(dSGg2)
  call comlib_sumcast(dSGv2)
#endif
#endif

#ifdef _RGGAUGE
  dSGg=(gauge%beta/6.0_DP)*(gauge%ct *gauge%cRG(0)*dSGg   &
 &                         +          gauge%cRG(1)*dSGg1  &
 &                     +2*gauge%ct_rg*gauge%cRG(1)*dSGg2)
  dSGv=(gauge%beta/6.0_DP)*(gauge%ct *gauge%cRG(0)*dSGv   &
 &                         +          gauge%cRG(1)*dSGv1  &
 &                     +2*gauge%ct_rg*gauge%cRG(1)*dSGv2)
#else
  dSGg=(gauge%beta/6.0_DP)*gauge%ct*dSGg
  dSGv=(gauge%beta/6.0_DP)*gauge%ct*dSGv
#endif

  this%gSF2inv= dSGg/knorm
  this%vSF    =-dSGv/knorm

  itraj = get_trajectory_number(this%status)
  if (0==nodeid) then
    write(str,'(I6,9E24.15)')itraj,this%gSF2inv,this%vSF,dSGg,-dSGv,knorm
    call print(this%logfile_gsf,TRIM(str))
  endif

  return
end subroutine

end module
